/* $Id: AppLogic.java 816 2009-11-17 14:52:52Z vasekremes $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.cashdesk;

import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.SOFAServer;
import org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.dev.IReceiptPrinter;
import org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui.IEvents;
import org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui.IRequests;
import org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui.PaymentRefusedException;
import org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui.PaymentType;
import org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui.SaleItemDetail;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Sale;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.SaleItem;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.StockItem;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.stock.IListStock;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.stock.ISale;

/**
  * 
  * @author Ondrej Cerny
  *
  */
public class AppLogic implements SOFAServer, SOFAClient, IEvents {
	
	// constants with interface names:
	private static final String IEVENTS_NAME = "guiEvents";
	private static final String IREQUESTS_NAME = "guiRequests";
	private static final String ISALE_NAME = "saleRegistration";
	private static final String ILISTSTOCK_NAME = "stockListing";
	private static final String IPRINT_RECEIPT = "printReceipt";
	
	/** Logger. */
	private static Logger logger = Logger.getLogger(AppLogic.class);
	
	/** GUI request listener. */
	private IRequests guiListener = null;
	
	/** Stock listing. */
	private IListStock stockLister = null;
	
	/** Sale registration handler. */
	private ISale saleRegistry = null;
	
	/** Receipt printer */
	private IReceiptPrinter receiptPrinter = null;
	
	/** Current sale. */
	private Sale currentSale = null;
	
	/** Indicates that cash desk number has been initialized. */
	private boolean isCashDeskNumber = false;
	
	/** Cash desk number. */
	private int cashDeskNumber = 1;
	
	/** Indicates wheter the sale has been already payed. */
	private volatile boolean isPayed = false;
	
	/** Indicates whether there is a payment in progress currently. */
	private volatile boolean paymentInProgress = false;
			
	// Private methods:
	/** Creates sale item detail from stock item and amount.
	  *
	  * @param si stock item
	  * @param amount amount
	  * @return sale item detail
	  */
	private SaleItemDetail createSaleItemDetailFromStockItemAndAmount(
			StockItem si, double amount) {
		SaleItemDetail result = null;
		if (si != null) {
			Product p = si.getProduct();
			if (p == null) {
				String ean = si.getEAN();
				logger.trace("Product of stock item with EAN='" +
						((ean != null) ? ean : "") + "' is null.");
			}
			result = new SaleItemDetailImpl(
					si.getEAN(),
					(p != null) ? p.getName() : "",
					(p != null) ? p.getDescription() : "",
					amount,
					si.getAmount(),
					si.getPrice());
		} else {
			logger.trace("Null stock item.");
		}
		return result;
	}
	
	/** Creates sale item from stock item and amount.
	  * 
	  * @param si stock item
	  * @param amount sale amount
	  * @return sale item
	  */
	private SaleItem createSaleItemFromStockItemAndAmount(
			StockItem si, double amount) {
		SaleItem result = null;
		if (si != null) {
			result = new SaleItem();
			result.setProduct(si.getProduct());
			result.setStockItemEAN(si.getEAN());
			result.setAmount(amount);
			result.setPricePerUnit(si.getPrice());
			result.setTotalPrice(amount * si.getPrice());
			if (si.getAmount() < result.getAmount()) {
				String ean = si.getEAN();
				logger.warn("Creating sale item: sale amount = " +
						Double.toString(result.getAmount()) + " > " +
						Double.toString(si.getAmount()) + " = stock amount; " +
						"EAN='" + ((ean != null) ? ean : "") + "'.");
			}
		} else {
			logger.trace("Null stock item when creating sale item.");
		}
		return result;
		
	}
			
	/** Adds sale item created from stock item and amount to current sale
	  * and returns sale item detail.
	  * 
	  * @param si stock item
	  * @param amount sale amount
	  */
	private SaleItemDetail addSaleItemAndReturnDetail(StockItem si, double amount) {
		SaleItemDetail result = null;
		if (si != null && currentSale != null) {
			SaleItem saleItem = 
				createSaleItemFromStockItemAndAmount(si, amount);
			result = 
				createSaleItemDetailFromStockItemAndAmount(si, amount);
			if (saleItem != null && result != null) {
				List<SaleItem> saleItems = currentSale.getSaleItems();
				if (saleItems == null) {
					saleItems = new LinkedList<SaleItem>();
				}
				saleItems.add(saleItem);
				currentSale.setSaleItems(saleItems);
				currentSale.setTotal(
						currentSale.getTotal() + saleItem.getTotalPrice());
			} else {
				result = null;
			}
		} else {
			logger.error("Null stock item or current sale " + 
					"when adding sale item.");
		}
		return result;
		
	}

	/** Returns cash desk number. */
	private int getCashDeskNumber() {
		if (!isCashDeskNumber) {
			// TODO: find out actual cash desk number from properties
		}
		return cashDeskNumber;
	}

	// IEvents methods:
	
	@Override
	/** Cancels current payment.
	  *
	  * @return <code>true</code> <i>iff</i> it is possible to cancel
	  * 
	  */
	public boolean cancelPayment() {
		logger.trace("Payment " + (paymentInProgress ? " cancel impossible." : " cancel."));
		return (!paymentInProgress);
	}

	/** Cancels current sale.
	  * 
	  * @returns <code>true</code> <i>iff</i> it is possible to cancel
	  * @see IEvents#cancelSale()
	  */
	@Override
	public boolean cancelSale() {
		if (isPayed) {
			logger.info("Sale has been alreayd payed, unable to cancel.");
			return false;
		}
		logger.info("Sale canceled.");
		currentSale = null;
		return true;
	}

	/** Returns current total to be payed.
	  * 
	  * @return total
	  * @see IEvents#getSum()
	  */
	@Override
	public double getSum() {
		if (isPayed || currentSale == null) {
			logger.trace("Returning total 0.00 because current sale " +
					"has been already payed or is null.");
			return 0;
		} else {
			double result = currentSale.getTotal();
			logger.trace("Returning total " + Double.toString(result) + ".");
			return result;
		}
	}

	/** Returns sale item detail.
	  * 
	  * @param EAN unique stock item identification
	  * @param amount amount
	  * @return sale item detail
	  * @see IEvents#itemSold(String, double) 
	  */
	@Override
	public SaleItemDetail itemSold(String EAN, double amount) {
		SaleItemDetail result = null;
		logger.trace("itemSold(" +
				((EAN != null) ? ("\"" + EAN + "\"") : ("null")) + 
				", " + Double.toString(amount) + ")");
		
		if (currentSale != null) {
			logger.info("Item sold: EAN='" + EAN + "', amount='" +
					Double.toString(amount) + "'.");
			if (stockLister != null) {
				StockItem si = null;
				// list stock and create SaleItemDetail 
				try {
					si = stockLister.getStockItemByEAN(EAN);
					result =
						addSaleItemAndReturnDetail(si, amount);
				} catch (InventoryUnreachableException iue) {
					logger.error("itemSold: inventory unreachable:" + iue.getMessage());
				}
				
			} else {
				logger.error("itemSold: null stock lister.");
			}
		} else {
		  logger.error("Item sold but current sale is null.");	
		}
		return result;
	}

	/** Registers current sale.
	  *
	  * @see IEvents#saleClosed()
	  */
	@Override
	public void saleClosed() {
		logger.trace("Closing sale.");
		if (currentSale != null) {
			currentSale.setClose(new Date());
			try {
				System.out.println("saleRegistry: " + saleRegistry);
				saleRegistry.registerSale(currentSale);
				logger.trace("Sale registered.");
				
				String receipt = "";
				
				List<SaleItem> sis = currentSale.getSaleItems();
				System.out.println("sis: " + sis);
				for(SaleItem si : sis) {
					System.out.println("si: " + si);
					Product p = si.getProduct();
					System.out.println("product: " + p);
					receipt += si.getAmount() +"x" + p.getName();
					
					receipt += "\t\t = " + si.getTotalPrice() + "\n";
				}
				
				receipt += "\n\n=====\nTOTAL\t\t" + currentSale.getTotal();
				System.out.println(receipt);
				receiptPrinter.printReceipt(receipt);
				
			} catch (Throwable t) {
				logger.error("Registering sale failed: " + t.getMessage());
			}
			currentSale = null;
		}
	}

	/** Updates sale item.
	  *
	  * <b>Not implemented yet.</b>
	  *
	  * @param EAN unique identification
	  * @param amount new amount
	  * @see IEvents#saleItemUpdated(String, double)
	  */
	@Override
	public void saleItemUpdated(String EAN, double amount) {
		// TODO implement this method
		logger.trace("Sale item update not implemented yet.");
		throw new InternalError(
				"AppLogic.saleItemUpdated() not implemented yet.");
	}

	/** Initializes current sale.
	  *
	  * @see IEvents#saleStarted() 
	  */
	@Override
	public void saleStarted() {
		isPayed = false;
		currentSale = new Sale();
		currentSale.setTotal(0);
		currentSale.setCashDeskNo(getCashDeskNumber());
		logger.trace("Sale started.");
	}

	/** Starts payment.
	  *
	  * @param pt payment type
	  * @param money payment sum
	  * @param cash given sum (makes sense for cash or cash-back payment)
	  * @return rest 
	  * @see IEvents#startPayment(org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui.IEvents.PaymentType, double, double)
	  */
	@Override
	public boolean startPayment(PaymentType pt, double money, double cash)
			throws PaymentRefusedException {
		
		logger.trace("Payment started.");
		paymentInProgress = true;
		try {
			
			if (currentSale != null) {
				if (money < currentSale.getTotal()) {
					logger.trace("Partial payment refused.");
					throw new PaymentRefusedException("Partial payment is not supported.");
				}
						
				if (pt == PaymentType.CASH) {
					logger.trace("Cash payment started.");
					// TODO: open cash box and wait until it is closed
					logger.debug("Open cash box and wait until it is closed.");
				} else if (pt == PaymentType.CARD) {
					logger.trace("Card payment started.");				
					if (money > currentSale.getTotal()) {
						logger.trace("Cash-back card payment not supported.");
						throw new PaymentRefusedException("Cash-back payment not allowed.");
					}

					// TODO: use card reader...
					logger.debug("Use card reader...payment refused.");
					throw new PaymentRefusedException("Unable to read card number.");

				} else {
					logger.error("Unsupported payment type.");
					throw new PaymentRefusedException("Unsupported payment type.");
				}
						
			} else {
				logger.error("Payment started but current sale is null.");
				throw new PaymentRefusedException("Payment out of sale process.");
			}
			isPayed = true;
			
		} finally {
			paymentInProgress = false;
			logger.trace("Payment finished.");
		}
		return false;
	}

	// SOFAServer method:
	/** Returns provided interface implementation. 
	  * 
	  * @param name interface name
	  * @return interface implementation
	  */
	@Override
	public Object getProvided(String name) {
		if (name != null) {
			if (name.equals(IEVENTS_NAME)) {
				return this;
			} else {
				logger.error("getProvided: unknown interface '" + name + "'.");
			}
		} else {
			logger.error("getProvided: null interface name.");
		}
		return null;
	}

    // SOFAClient method:
	/** Sets required interface.
	  *  
	  * @param name interface name
	  * @param impl implementing object
	  */
	@Override
	public void setRequired(String name, Object impl) {
		if (name != null && impl != null) {
			if (name.equals(ILISTSTOCK_NAME)) {
				if (impl instanceof IListStock) {
					stockLister = (IListStock) impl;
				} else {
					logger.error("setRequired: '" + ILISTSTOCK_NAME + "'" +
							"does not implement IListStock interface.");
				}
			} else if (name.equals(ISALE_NAME)) {
				if (impl instanceof ISale) {
					saleRegistry = (ISale) impl;
				} else {
					logger.error("setRequired: '" + ISALE_NAME + "'" +
							"does not implement ISale interface.");
				}
			} else if (name.equals(IREQUESTS_NAME)) {
				if (impl instanceof IRequests) {
					guiListener = (IRequests) impl;
				} else {
					logger.error("setRequired: '" + IREQUESTS_NAME + "' " +
							"does not implement IRequests interface.");
				}
			} else if (name.equals(IPRINT_RECEIPT)) {
				if (impl instanceof IReceiptPrinter) {
					receiptPrinter = (IReceiptPrinter) impl;
				} else {
					logger.error("setRequired: '" + IPRINT_RECEIPT + "' " +
							"does not implement IReceiptPrinter interface");
				}
			} else {
				logger.error("setRequired: unkonw interface name '" +
						name +"'");
			}
		} else {
			logger.error("setRequired: null name or object.");
		}
	}
}

