/* $Id: Shell.java 813 2009-11-17 12:05:05Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.cashdesk.gui;

import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import org.objectweb.dsrg.sofa.SOFALifecycle;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.SOFAServer;

/**
  * @author ondra
  *
  */
public class Shell implements SOFAClient, SOFALifecycle, SOFAServer {

  private final static String IEVENTS_NAME = "events";
  private final static String IREQUESTS_NAME = "requests";
  
  /** GUI components handler. */
  SwingInitializer componentsHandler = null;
  /** IRequest calls handler. */
  RequestHandler requestHandler = null;
  /** Indicates whether the UI has been started. */
  private volatile boolean started = false;
	
	/** Constructor. */
	public Shell() {
		componentsHandler = new SwingInitializer();
	}
	
	/** Returns request handler.
	  *
	  * @return IRequest calls handler
	  */
	private synchronized RequestHandler getReqHandler() {
		if (requestHandler == null) {
			requestHandler = new RequestHandler(componentsHandler);
		}
		return requestHandler;
	}

	// SOFALifeCycle methods:
	/** Starts the component.
	  * 
	  * <p>Enqueues starting the GUI.</p>
	  */
	public synchronized void start() {
		if (!started) {
			started = true;
			SwingUtilities.invokeLater(componentsHandler);
		}
	}
	
	/** Notifies that the component will be stopped nearly.
	  *
	  * <p>Enqueues stopping the GUI. </p>
	  */
	public void stopping() {
		SwingUtilities.invokeLater(componentsHandler);
	}
	
	/** Stops the component.
	  * 
	  */
	public synchronized void stop() {
		if (started) {
			// TODO: exit the GUI
			started = false;
		}
	}

  // SOFAClient method
  /** Registers required interface.
    *
    * @param name name of the interface
    * @param iface interface implementation (or connector)
    */
  public void setRequired(String name, Object iface) {
    if (name != null && name.equals(IEVENTS_NAME)) {
      if (iface != null && iface instanceof IEvents) {
        try {
          ControllerHelper.initializeOnlyOnce((IEvents)iface);
        } catch (Throwable t) {
          System.err.printf("Warning: Shell.setRequired(\"%s\", ...):" +
            "ControllerHelper.initializeOnlyOnce(...) failed with " +
            " message: %s\n", name, t.getMessage());
        }
      } else {
        System.err.printf("Error: interface \"%s\" does not implement " +
          "IEvents!\n", name);
      }
    }
    // ignore others ...
  }

  // SOFAServer method
  /** Returns provided interface.
    *
    * @param name name of the interface
    * @return interface implementation
    */
  public Object getProvided(String name) {
    if (name != null) {
      if (name.equals(IREQUESTS_NAME)) {
    	return getReqHandler();
      } else {
        throw new IllegalArgumentException("Unkown provided interface '" + name + "'.");
      }
    } else {
      throw new NullPointerException("Unable to return provided interface with null name.");
    }
  }
	
	/* For debugging. */
	public static void main(String[] args) {
		
		Shell inst = new Shell();
		try {
			UIManager.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");
		} catch (Exception e) {
			System.err.printf("Warning: Unable to set look and feel: %s\n", e.getMessage());
		}
		SwingUtilities.invokeLater(inst.componentsHandler);
	}

}
