/* $Id: ITransact.java 652 2009-08-24 00:17:16Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db;

import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.Objects;

/** Provides transactions over object storage.
  *
  * <p>
  * We assume automatic transaction start. From the view of
  * single caller's context (thread), we may define following
  * conditions:<br/>
  * <ul><li>each method call belongs to a transaction</li>
  *     <li>lets say <code>commit</code> and <code>rollback</code>
  *         are transcation <i>delimiters</i>, then all calls between two
  *         successive delimiters belong to same transaction as the second
  *         delimiter.</li>
  *     <li>transaction may be <i>aborted</i> at any time by the database,
  *         then next commit will fail (no changes will be actually written)</li>
  * </ul>
  * </p>
  *
  * @author Ondrej Cerny
  */
public interface ITransact {
  /** Fetches objects matching query <code>q</code> from database.
    *
    * @param q query
    * @return objects matching query
    * @throws DatabaseUnreachableException database is unreachable
    * @throws TransactionAbortedException transaction has been aborted
    * @throws InvalidContextException invalid context
    *   (should be handle similarly as TransactionAbortedException)
    */
  public <T extends AbstractBase> Objects<T> fetch(IQuery i)
  throws DatabaseUnreachableException,
         TransactionAbortedException,
         InvalidContextException;

  /** Fetches objects matching query <code>q</code> from database.
    *
    * @param q query
    * @param startIndex index of the first returned object
    * @param endIndex index of the first object behind the last returned object
    * @return objects matching query
    * @throws DatabaseUnreachableException database is unreachable
    * @throws TransactionAbortedException transaction has been aborted
    * @throws InvalidContextException invalid context
    *   (should be handle similarly as TransactionAbortedException)
    */
  public <T extends AbstractBase> Objects<T> fetch(IQuery q,
                                                   int startIndex,
                                                   int endIndex)
  throws DatabaseUnreachableException,
         TransactionAbortedException,
         InvalidContextException;

  /** Stores object onto database.
    * 
    * @param object an object to be stored
    * @return stored object
    * @throws DatabaseUnreachableException database is unreachable
    * @throws TransactionAbortedException transaction has been aborted
    * @throws InvalidContextException invalid context
    *   (should be handle similarly as TransactionAbortedException)
    */
  public <T extends AbstractBase> T store(T object)
  throws TransactionAbortedException, 
         DatabaseUnreachableException,
         InvalidContextException;

  /** Commits (unrevertibly writes) changes to databse.
    *
    * @return false if and only if the transaction has been aborted 
    * @throws DatabaseUnreachableException database is unreachable
    */ 
  public boolean commit()
  throws DatabaseUnreachableException;

  /** Rolls changes back.
    *
    * @return false if and only if the transaction has been aborted 
    * @throws DatabaseUnreachableException database is unreachable
    */
  public void rollback()
  throws DatabaseUnreachableException;

  /** Indicates whether ABORT has occured within this transaction.
    *
    * @return true if and only if the transaction has been aborted
    */
  public boolean isAborted();

}
