/* $Id: InternalObjectIdManager.java 806 2009-11-15 10:08:58Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db.storage;

import java.math.BigDecimal;
import java.util.Collection;

import org.neodatis.odb.ODB;
import org.neodatis.odb.ObjectValues;
import org.neodatis.odb.Objects;
import org.neodatis.odb.Values;
import org.neodatis.odb.core.query.IValuesQuery;
import org.neodatis.odb.impl.core.query.values.ValuesCriteriaQuery;
import org.apache.log4j.Logger;


/** Manages internal object ids.
  *
  * @author Ondrej Cerny
  */
public class InternalObjectIdManager {
  /** Represents state of object ID manager from the view of database. */
  public class StateDescription {
    private int lastId;
    private boolean isDirty;
    public StateDescription() {
      lastId = 1;
      isDirty = false;
    }
    public void setId(int id) {
      lastId = id;
    }
    public int getId() {
      return lastId;
    }
    public void setDirty(boolean d) {
      isDirty = d;
    }
    public boolean isDirty() {
      return isDirty;
    }
  }
  /** Logger */
  private static Logger logger = Logger.getLogger(InternalObjectIdManager.class.getName());
  /** Id. */
  private int lastAssignedId;
  
  /** Recovers the {@link StateDescription}.
    *
    * <p>The recovery means that we have to find
    *    maximal internal_object_ID field out
    *    (all around the database).
    * </p>
    * 
    * @param session
    * @param managedClasses
    * @param sd
    * @return <code>true</code> on success
    */
  private boolean recoveryProcedure(ODB session, Collection<Class<?>> managedClasses,
		  StateDescription sd) {
	int maxID = sd.getId();
	int steps = 0;
    for(Class<?> clazz : managedClasses) {
    	++steps;
    	logger.info("Processing step " + steps + ".");
    	IValuesQuery vq = new ValuesCriteriaQuery(clazz).max("internal_object_ID", "ID");
    	Values v = session.getValues(vq);
    	ObjectValues ov = v != null ? v.getFirst() : null;
    	Object ovo = ov != null ? ov.getByAlias("ID") : null;
    	BigDecimal aggregate = ovo != null && ovo instanceof BigDecimal ? ((BigDecimal)ovo) : null;
    	if (aggregate != null) {
    		if (aggregate.intValue() > maxID) {
    			maxID = aggregate.intValue();
    			logger.info("Step " + steps + ": increasing OID to " + maxID); 
    		}
    	} else {
    		logger.warn("Step " + steps + " failed. [" + clazz.getName() + "]");
    		return false;
    	}
    }
    logger.info("Proccess finished with OID " + maxID);
    sd.setId(maxID);
    return true;
  }

  /** Fetches the state description from database.
    *
    * @param session
    * @param managedClasses collection of classes
    * (may be <code>null</code> if and only if <code>fetchDirty == true</code>)
    * @param fetchDirty
    * @return new or fetched state description on success,
    *  null on failures
    */
  private StateDescription fetchStateDesc(ODB session, Collection<Class<?>> managedClasses,
		  boolean fetchDirty) {
    StateDescription result = null;
    try {
      Objects<StateDescription> states = session.getObjects(
        StateDescription.class);
      if (states == null || states.size() == 0) {
        logger.trace("No state descriptions in the database. " +
          "This is probably the first run.\n");
        result = new StateDescription();
      } else if (states.size() == 1) {
        StateDescription s = states.getFirst();
        if (s == null) {
          logger.error("The database has probably crashed. " +
                  "Automatic rocovery is not possible. " +
                  "Stop application, perform manual database cleanup " +
                  "and start the application again, please.");
        } else if (s.isDirty && !fetchDirty) {
          logger.warn("The database has probably not been closed properly. " +
            "Starting recovery procedure.");
          if (recoveryProcedure(session, managedClasses, s)) {
        	result = s;
        	logger.warn("Recovery successfull.");
          } else {
          	logger.warn("Recovery failed. Stop application, " +
          			"perform manual database cleanup " +
                    "and start the application again, please.");  
          }
        } else {
          result = s;
          logger.trace("Loaded ID=" + lastAssignedId + ".");
        }
      } else {
        logger.error("The database contains multiple state descriptions. " +
          "Stop the application, clean reduntant state descriptions up " +
          "and start the aplication again, please.");
      }
    } catch (Exception e) {
      logger.fatal("System error: Unable to fetch state description.");
      result = null;
    }
    return result;
  }

  /** Fetches lastly assigned id from dbase.
    * 
    * Calls of init() and close() must form correct bracketing.
    *
    * @param session correctly initialized database session,
    *   this method calls commit on the session before return
    * @return true if and only if the initialization was successfull
    */
  public synchronized boolean init(ODB session, Collection<Class<?>> managedClasses) {
    boolean initialized = false;
    lastAssignedId = 1;
    StateDescription s = fetchStateDesc(session, managedClasses, false);
    try {
      if (s != null) {
        lastAssignedId = s.getId();
        s.setDirty(true);
        session.store(s);
        initialized = true;
      }
      session.commit();
    } catch (Exception e) {
      logger.fatal("System error: Unable to fetch state description.");
      initialized = false;
    }
    return initialized;
  }

  /** Stores lastly assigned id to the database.
    *
    * Calls of init() and close() must form correct bracketing.
    *
    * @param session correctly initialized database session,
    *   after this method calls commit on the session before
    *   return
    */
  public synchronized void close(ODB session) {
    StateDescription s = fetchStateDesc(session, null, true);
    if (s != null) {
      s.setDirty(false);
      s.setId(lastAssignedId);
      session.store(s);
      session.commit();
      logger.trace("Saved ID=" + lastAssignedId + ".");
    } else {
      // Error!
      throw new RuntimeException("Unable to get StateDescription for update " +
        "at InternalObjectIdManager.close.");
    }
  }

  /** Incerments id and returns.
    * 
    * @return next id
    */
  public synchronized int getNextId() {
    try {
      ++lastAssignedId;
    } catch (ArithmeticException e) {
      logger.fatal("Internal id overflown!");
      // TODO: we should somehow notify the ODBWrapper that the database is
      //       now in error state and it would stop itself
      // note that this exception will be ignored since this method is called
      // from database trigger context
      throw e;
    }
    return lastAssignedId;
  }
}
