/* $Id: ODBSessionAdapter.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db.storage;

import java.util.Set;
import java.util.Collection;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.ODB;
import org.neodatis.odb.OID;
import org.neodatis.odb.ODBServer;
import org.neodatis.odb.Objects;
import org.apache.log4j.Logger;

/** Adapts ODB session/connection to be usable as
  * ObjectStorage session. 
  *
  * @author Ondrej Cerny
  */
public class ODBSessionAdapter {

  /** Logger. */
  private static Logger logger = Logger.getLogger(ODBSessionAdapter.class);

  /** ODB session. */
  private ODB odbSession;

  /** Client-side insert trigger that adds internal id to inserted objects.*/
  private ClientInsertTrigger cit;

  /** Client-side select trigger that allows returning set of fetched ids. */
  private ClientSelectTrigger cst;

  /** Creates new ODBSessionAdapter. */
  public ODBSessionAdapter() {
    logger.trace("Creating new ODBSessionAdapter instance.");
    odbSession = null;
    cit = null;
    cst = null;
  }

  /** Activates session.
    *
    * @param server ODB server
    * @param classes classes that should be covered by triggers
    * @param returnsOIDs should this session controll fetched object ids
    */
  public void activate(ODBServer server, Collection<Class<?>> classes, boolean returnsOIDs) {
    boolean newSession = false;
    // check existence and compatibility of required and current sessions
    if (odbSession != null && this.cst != null && !returnsOIDs) {
      logger.trace("Closing old ODB session because of incompatibility at " +
        "ODBSessionAdapter@" + this.hashCode() + ".");
      odbSession.close();
      odbSession = null;
    }
    // open new session if neccessary
    if (odbSession == null) {
      logger.trace("Opening new ODB session at " + 
        "ODBSessionAdapter@" + this.hashCode() + ".");
      newSession = true;
      odbSession = server.openClient(ODBWrapper.DATABASE_NAME);
      // add insert trigger
      logger.trace("Adding insert trigger at " +
        "ODBSessionAdapter@" + this.hashCode() + ".");
      if (cit == null) {
        cit = new ClientInsertTrigger();
      }
      if (classes != null) {
        for(Class<?> clazz : classes) {
          odbSession.addInsertTrigger(clazz, cit);
        }
      }
    }
    // upgrade session with select trigger if neccessary
    if (returnsOIDs && (this.cst == null || newSession)) {
      logger.trace("Adding select trigger at " +
        "ODBSessionAdapter@" + this.hashCode() + ".");
      if (cst == null) {
        cst = new ClientSelectTrigger();
      }
      if (classes != null) {
        for (Class<?> clazz: classes) {
          odbSession.addSelectTrigger(clazz, cst);
        }
      }
    }
  }

  /** Deactivates the session. 
    */ 
  public void deactivate() {
    // TODO: try to reuse sessions
    //
    // to be reusable, the session must be correctly commited
    // and in valid state...
    if (odbSession != null) {
      if (!odbSession.isClosed()) {
        logger.trace("Closing ODB session at " +
          "ODBSessionAdapter@" + this.hashCode() + ".");
        odbSession.close();
      }
      odbSession = null;
    }
  }

  /** Returns true iff fetch calls on this session will return OIDs.
    *
    * @return true iff fetch returns OIDs
    */ 
  public boolean willReturnOIDs() {
    return (cst != null);
  }

  /** Fetches objects matching query.
    *
    * @param q query
    * @param startIndex index of the first returned object
    * @param endIndex index of the first object behind the last returned object
    * @return object representing query result
    */
  public <T extends AbstractBase> QueryResult<T> fetchObjects(IQuery q,
                                                              int startIndex,
                                                              int endIndex) {
    //TODO: database itself does not handle the case when the range of fetched
    //      objects is limited and there is an orderBy clause - it order just
    //      objects within the range
    //
    //      we must handle this case ourselves - fetch all objects with orderBy
    //      clause an then shirnk the range, shrink the set of OIDs and
    //      disconnect objects that will not be returned
    Objects<T> result = odbSession.getObjects(q, false, startIndex, endIndex);
    Set<OID> resultOIDs = null;
    if (cst != null) {
      resultOIDs = cst.getSetAndClear();
    }
    return new QueryResultImpl<T>(result, resultOIDs);
  }

  /** Disconnect objects related to OIDs.
    * 
    * @param setOfIds set of OIDs.
    */
  public void disconnectObjects(Set<OID> setOfIds) {
    // NOT TESTED YET!
    if (setOfIds != null) {
      for(OID oid : setOfIds) {
        Object o = odbSession.getObjectFromId(oid);
        if (o != null) {
          odbSession.disconnect(o);
        }
      }
    }
  }

  /** Stores object to database.
    *
    * @param object object to be stored
    */
  public OID storeObject(AbstractBase object) {
    return odbSession.store(object);
  }

  /** Commit changes. */
  public void commit() {
    odbSession.commit();
  }

  /** Rollback changes. */
  public void rollback() {
    odbSession.rollback();
  }

}
