/* $Id: ODBWrapper.java 806 2009-11-15 10:08:58Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db.storage;

import java.util.Properties;
import java.util.Set;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.OID;
import org.neodatis.odb.ODB;
import org.neodatis.odb.ODBServer;
import org.neodatis.odb.ODBFactory;
import org.neodatis.odb.OdbConfiguration;
import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFALifecycle;
import org.objectweb.dsrg.sofa.SOFAParametrized;

public class ODBWrapper implements ObjectStorage, SOFALifecycle, SOFAParametrized {

  /** Database name. */
  public static final String DATABASE_NAME = "base";
  private static final String DATABASE_FILENAME_PROPERTY_NAME = "database.file.name";
  private static final int MAX_ODBENGINE_START_ATTEMPTS = 1000;
  private static final int DYNPORT_MIN = 49152;
  private static final int DYNPORT_MAX = 65535;

  /** Manages object ids for objects remoting. */
  public static InternalObjectIdManager internalObjectIdManager =
    new InternalObjectIdManager();
  /** Indicates whether this component may open sessions. */
  private boolean mayOpen = false;
  /** Indicates whether initialization of component has finished (either successfully or not). */
  private boolean initialized = false;
  /** Contains ODB engine (server). */
  private ODBServer server = null;
  /** Manages database sessions. */
  private SessionPool sessionPool = null;
  /** Logger. */
  private static Logger logger = Logger.getLogger(ODBWrapper.class);
  /** Database file name. */
  private String databaseFileName = System.getProperty("java.io.tmpdir") + java.io.File.separator + "sofashop.neodatis";


  /** Initializes this component.
    */
  private void init() {
	String tmpDatabaseFileName;
	synchronized (this) {
		tmpDatabaseFileName = new String(databaseFileName);
	}
    
    // result of server initialization
    boolean result = false;
    
    try {
    	//open database server
    	logger.trace("About to open database engine with file " + tmpDatabaseFileName + ".");
    	OdbConfiguration.setShareSameVmConnectionMultiThread(true);
    	int portNumber = 0; //  unix finds available port
    	int attempts = 0;
    	boolean giveUp = false;
    	java.util.Random rnd = null;
    	while (server == null && !giveUp) {
    		try {
    			logger.trace("About to open ODB database.");
    			server = ODBFactory.openServer(portNumber);
    			server.addBase(DATABASE_NAME, tmpDatabaseFileName);
    			logger.trace("ODB database engine started.");
    		} catch (Throwable t) {
    			server = null;
    			logger.trace("Starting database from file " +
    					tmpDatabaseFileName + " port " + portNumber + 
    					" failed (" + t.getClass().getName() + ": " +
    					t.getMessage() + "). ");
    			
    			// get random dynamic port number for next attempt
    			if (rnd == null) {
    				rnd = new java.util.Random();
    			}
    			portNumber = rnd.nextInt(DYNPORT_MAX - DYNPORT_MIN) + DYNPORT_MIN;
    			
    			++attempts;
    			if (attempts >= MAX_ODBENGINE_START_ATTEMPTS) {
    				throw t;
    			} else {
    				logger.trace("Retrying.");
    			}
    			
    		}
    	}
    	
    	if (server != null) {
    		//initialize id manager
    		ODB client = server.openClient(DATABASE_NAME);
    		result = internalObjectIdManager.init(client, AbstractBase.subClasses);
    		if (result) {
    			logger.info("Database \"" + tmpDatabaseFileName + "\" started");
    		} else {
    			logger.error("Database \"" + tmpDatabaseFileName + "\" startup failed.");    			
    		}
    		client.close();
    	
    		//create session pool
    		sessionPool = new SessionPool(server, AbstractBase.subClasses);
    	} else {
    		logger.error("Database \"" + tmpDatabaseFileName + "\" startup failed.");
    	}

    } catch (Throwable t ) {
    	result = false;
    	logger.error("Unable to initialize database because of: " +
    			t.getClass().getName() + ": " + t.getMessage());
    }   
    synchronized (this) {
        mayOpen = result;
        initialized = true;		
	}
  }

  /**
    * @see org.objectweb.dsrg.sofa.SOFARunnable#start()
    */
  @Override
  public void start() {
	logger.trace("start");	  
	init();
	
	boolean tmp = false;
	synchronized (this) {
	  tmp = mayOpen;
	}
	if (!tmp) {
      logger.warn("Initialization failed.");
	}
  }

  /**
    * @see org.objectweb.dsrg.sofa.SOFAStoppable#stopping()
    */
  @Override
  public void stopping() {
	logger.trace("stopping");
    mayOpen = false;
  }
  /**
    * 
    * @see org.objectweb.dsrg.sofa.SOFAStoppable#stop()
    */
  @Override
  public void stop() {
	logger.trace("stop");  
    // stop all opened sessions
    sessionPool.forceCloseAll();
    try {
      ODB client = server.openClient(DATABASE_NAME);
      internalObjectIdManager.close(client);
      client.close();
      logger.info("Database \"" + databaseFileName + "\" stopped.");    			

    } catch (Exception e) {
      throw new RuntimeException("Unable to save InternalObjectIdManager's state.", e);
    }
    try {
      if (server != null) {
        server.close();
      }
    } catch (Exception e) {
      throw new RuntimeException("Unable to close database.", e);
    }
    logger.trace("stopped");
  }
  /**
    * 
    * @see org.objectweb.dsrg.sofa.SOFAParametrized#setProperties(java.util.Properties)
    */
  @Override
  public void setProperties(Properties arg0) {
    logger.trace("setProperties");
    if (arg0 != null) {
    	logger.trace("getting property " + DATABASE_FILENAME_PROPERTY_NAME);
		String prop = arg0.getProperty(DATABASE_FILENAME_PROPERTY_NAME);
		if (prop != null && prop.length() > 0) {
			synchronized (this) {
    			databaseFileName = new String(prop);
    		}
    	} else {
    		logger.info("Property '" + DATABASE_FILENAME_PROPERTY_NAME + "' " +
    				"has not been set. Using default value " +
    				"'"+ databaseFileName + "'.");
    	}
    }
  }
  

  // ObjectStorage methods:
  /**
    * 
    */
  @Override
  public int open(boolean verbose) {
	// wait while not initialized
	boolean pass = false;
	while (!pass) {
		synchronized (this) {
			pass = initialized;
		}
		if (!pass) {
			Thread.yield();
		}
	}
	// create session
    return (mayOpen) ? sessionPool.createActiveSession(verbose) : -1;
  }
  public <T extends AbstractBase> QueryResult<T> fetchObjects(
                                                   int transactionId,
                                                   IQuery q,
                                                   int startIndex,
                                                   int endIndex) {
    ODBSessionAdapter osa = sessionPool.getSession(transactionId);
    if (osa != null) {
      return osa.fetchObjects(q, startIndex, endIndex);
    } 
    return null;
  }
  /**
    * 
    * @see org.objectweb.dsrg.sofa.demos.sofashop.db.storage.ObjectStorage#storeObject(int, org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase)
    */
  @Override
  public <T extends AbstractBase> T storeObject(int transactionId, T object) {
    ODBSessionAdapter osa = sessionPool.getSession(transactionId);
    if (osa != null) {
      OID res = osa.storeObject(object);
      if (res != null) {
        return object;
      }
    } 
    return null;
  }
  /**
    * 
    * @see org.objectweb.dsrg.sofa.demos.sofashop.db.storage.ObjectStorage#disconnectObjects(int, java.util.Set)
    */
  @Override
  public void disconnectObjects(int transactionId, Set<OID> setOfIds) {
    ODBSessionAdapter osa = sessionPool.getSession(transactionId);
    if (osa != null) {
      osa.disconnectObjects(setOfIds);
    } 
  }
  /**
    * 
    * @see org.objectweb.dsrg.sofa.demos.sofashop.db.storage.ObjectStorage#commitAndClose(int)
    */
  @Override
  public void commitAndClose(int transactionId) {
    ODBSessionAdapter osa = sessionPool.getSession(transactionId);
    if (osa != null) {
      osa.commit();
      sessionPool.closeSession(transactionId);
    } 
  }
  /**
    *  
    * @see org.objectweb.dsrg.sofa.demos.sofashop.db.storage.ObjectStorage#rollbackAndClose(int)
    */
  @Override
  public void rollbackAndClose(int transactionId) {
    ODBSessionAdapter osa = sessionPool.getSession(transactionId);
    if (osa != null) {
      osa.rollback();
      sessionPool.closeSession(transactionId);
    } 
  }
}
