/* $Id: SessionPool.java 644 2009-08-19 16:07:34Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db.storage;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Collection;
import org.neodatis.odb.ODB;
import org.neodatis.odb.ODBServer;
import org.apache.log4j.Logger;

/** Opens and manages database sessions.
  *
  * @author Ondrej Cerny
  */
public class SessionPool {
  /** Logger. */
  private static Logger logger = Logger.getLogger(SessionPool.class);
  /** Classes managed by the database. */
  private Collection<Class<?>> managedClasses;
  /** Maps session id to the ODB session. */
  Map<Integer, ODBSessionAdapter> activeSessions;
  /** Retains deactivated plain sessions that may be reactivated. */
  List<ODBSessionAdapter> deactivatedPlainSessions;
  /** Retains deactivated verbose sessions that may be reactivated. */
  List<ODBSessionAdapter> deactivatedVerboseSessions;
  /** Provides sessions. */
  ODBServer server;
  /** Counts sessions. */
  int sessionCounter;
  /** Indicates whether the session counter has overflown. */
  boolean overflown;
  /** Creates new pool attached to given server.
    *
    * @param server ODB server
    */
  public SessionPool(ODBServer server, Collection<Class<?>> managedClasses) {
    this.server = server;
    this.managedClasses = managedClasses;
    activeSessions = new HashMap<Integer, ODBSessionAdapter>();
    deactivatedPlainSessions = new LinkedList<ODBSessionAdapter>();
    deactivatedVerboseSessions = new LinkedList<ODBSessionAdapter>();
    sessionCounter = 0;
    overflown = false;
  }
  /** Puts ODB session adapter to list of deactivated sessions.
    *
    * @param odbsa deactivated session adapter
    */
  private void deacitvateSession(ODBSessionAdapter odbsa) {
    if (odbsa != null) {
      boolean verbose = odbsa.willReturnOIDs();
      odbsa.deactivate();
      if (verbose) {
        deactivatedVerboseSessions.add(odbsa);
      } else {
        deactivatedPlainSessions.add(odbsa);
      }
    }
  }
  /** Returns a ODB session from list of deactivated sessions.
    *
    * @return ODB session of null if the list is empty
    */
  private ODBSessionAdapter getDeactivatedOdbSession(boolean verbose) {
    List<ODBSessionAdapter> firstDList, secondDList;
    if (verbose) {
      firstDList = deactivatedVerboseSessions;
      secondDList = deactivatedPlainSessions;
    } else {
      firstDList = deactivatedPlainSessions;
      secondDList = deactivatedVerboseSessions;
    }
    ODBSessionAdapter result = null;
    if (firstDList != null && !firstDList.isEmpty()) {
      result = firstDList.remove(0);
    } else {
      if (secondDList != null && !secondDList.isEmpty()) {
        result = secondDList.remove(0);
      }
    }
    return result;
  }
  /** Returns next non-occupied integer.
    * 
    * @return next id
    */
  private int getNextSessionId() {
    int runs = 0;
    if (overflown) {
      try {
        ++sessionCounter;
      } catch (ArithmeticException e) {
        sessionCounter = 1;
      }
      while (activeSessions.containsKey(new Integer(sessionCounter))) {
        try {
          ++sessionCounter;
        } catch (ArithmeticException e) {
          ++runs;
          if (runs > 1) {
            throw e;
          } else {
            sessionCounter = 1;
          }
        }
      }
    } else {
      try {
        ++sessionCounter;
      } catch (ArithmeticException e) {
        overflown = true;
        return getNextSessionId();
      }
    }
    return sessionCounter;
  }
  /** Creates new active session.
    *
    * @param verboseSession verbose session returns OIDs of
    *   fetched objects as well
    * @return integral session id
    *
    * Internally it can reactivate previously closed session.
    *
    */
  public synchronized int createActiveSession(boolean verboseSession) {
    ODBSessionAdapter tmp = getDeactivatedOdbSession(verboseSession);
    int id = getNextSessionId();
    if (tmp == null) {
      tmp = new ODBSessionAdapter();
    } else {
      logger.trace("Reusing session for id " + id + ".");
    }
    tmp.activate(server, managedClasses, verboseSession);
    activeSessions.put(new Integer(id), tmp);
    return id;
  }
  /** Returns ODB session currently maintained under given session id.
    *
    * @param sessionId
    * @return ODB session or null (if the id is not currently active)
    */
  public synchronized ODBSessionAdapter getSession(int sessionId) {
    logger.trace("Fetching session for id " + sessionId + ".");
    ODBSessionAdapter tmp = activeSessions.get(sessionId);
    if (tmp == null) {
      logger.error("Session for id " + sessionId +
        " is not mapped! Returning NULL.");
    }
    return tmp;
  }
  /** Deactivates sessions. 
    * 
    * Currently it does not allow session reuse.
    *
    * @param sessionId id of the session
    */
  public synchronized void closeSession(int sessionId) {
    ODBSessionAdapter tmp = activeSessions.remove(sessionId);
    if (tmp != null) {
      tmp.deactivate();
      
    } else {
      logger.error("Attempt to close not mapped session id " +
        sessionId + ".");
    }
  }
  /** Forces closing of all sessions.
    *
    * <b>TODO</b>
    */
  public void forceCloseAll() {
    //TODO
  }
}
