/* $Id: MutexTransactionHelper.java 652 2009-08-24 00:17:16Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db.transaction;

import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.storage.ObjectStorage;
import org.objectweb.dsrg.sofa.demos.sofashop.db.storage.QueryResult;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.neodatis.odb.Objects;
import org.neodatis.odb.core.query.IQuery;
import org.apache.log4j.Logger;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

public class MutexTransactionHelper extends TransactionHelperBase
implements ITransact {
  /** Logger. */
  private static Logger logger = Logger.getLogger(MutexTransactionHelper.class);

  /** Semaphore... */
  private static volatile Semaphore sem = new Semaphore(1);

  /** Lock timeout (ms). */
  private static final int LOCK_TIMEOUT = 500;

  /** Maximal number of retries caused by interrupts for single attempt. */
  private static final int MAX_INTERRUPTS = 10;

  /** Aborted flag. */
  private boolean aborted;

  /** Creates new instance.
    *
    * @param o object storage
    */
  public MutexTransactionHelper(ObjectStorage o) {
    super(o, false);
    aborted = false;
  }

  /** Initializes database session.
    *
    * @return session ID
    */
  private int getActiveTransaction()
  throws DatabaseUnreachableException, TransactionAbortedException {
    if (aborted) {
      throw new TransactionAbortedException("Transaction already aborted.");
    }
    int result = -1;
    boolean isActive = isActive();
    if (!isActive) {
      // acquire exclusive access and try to activate session
      boolean lockresult = false;
      int interrupts = 0;
      while (true) {
        try {
          lockresult = sem.tryAcquire(1, LOCK_TIMEOUT, TimeUnit.MILLISECONDS);
          break;
        } catch (InterruptedException ie) {
          interrupts++;
          if (interrupts > MAX_INTERRUPTS) {
            logger.warn("Reached MAX_INTERRUPTS within single semaphore acquire attempt.");
            break;
          }
        }
      }
      if (!lockresult) {
        aborted = true;
        throw new TransactionAbortedException(
          "Unable to lock database access for transaction.");
      }
    }
    result = getTransactionNumber();
    if (result < 0) {
      if (!isActive) {
        sem.release(1);
      }
      aborted = true;
      logger.warn("Unable to open database session for transaction.");
      throw new DatabaseUnreachableException(
        "Unable to open database session for transaction.");
    }
    return result;
  }

  // ITransact methods:

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#fetch(IQuery) }
    * method.
    */ 
  public <T extends AbstractBase> Objects<T> fetch(IQuery q)
  throws DatabaseUnreachableException, TransactionAbortedException {
    return fetch(q, -1, -1);
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#fetch(IQuery, int, int) }
    * method.
    */ 
  public <T extends AbstractBase> Objects<T> fetch(IQuery q,
                                                   int startIndex,
                                                   int endIndex)
  throws DatabaseUnreachableException, TransactionAbortedException {
    QueryResult<T> qr = objectStorageProvider.fetchObjects(
                                                getActiveTransaction(),
                                                q, startIndex, endIndex);
    if (qr != null) {
      return qr.getObjects();
    } else {
      return null;
    }
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#store(T) }
    * method.
    */ 
  public <T extends AbstractBase> T store(T object)
  throws DatabaseUnreachableException, TransactionAbortedException {
    return objectStorageProvider.storeObject(getActiveTransaction(), object);
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#commit() }
    * method.
    */ 
  public boolean commit()
  throws DatabaseUnreachableException {
    boolean isActive = isActive();
    boolean result = (!aborted) && isActive;
    try {
      if (isActive) {
        objectStorageProvider.commitAndClose(getActiveTransaction());
      }
    } catch (TransactionAbortedException t) {
      result = false;
    } finally {
      endTransaction();
      if (!aborted && isActive) {
        sem.release(1);
      }
      aborted = false;
    }
    return result;
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#rollback() }
    * method.
    */ 
  public void rollback()
  throws DatabaseUnreachableException {
    boolean isActive = isActive();
    try {
      if (isActive) {
        objectStorageProvider.rollbackAndClose(getActiveTransaction());
      }
    } catch (TransactionAbortedException t) {
    } finally {
      endTransaction();
      if (!aborted && isActive) {
        sem.release(1);
      }
      aborted = false;
    }
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#isAborted() }
    * method.
    */ 
  public boolean isAborted() {
    return aborted;
  }
}
