/* $Id: TransactionManager.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.db.transaction;

import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.storage.ObjectStorage;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;
import org.apache.log4j.Logger;
import org.neodatis.odb.Objects;
import org.neodatis.odb.core.query.IQuery;
import org.objectweb.dsrg.sofa.SOFAServer;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.SOFAThreadHelper;
import java.util.Map;
import java.util.HashMap;

public class TransactionManager implements ISelect, ITransact, SOFAClient, SOFAServer {
  /** Logger. */
  private static Logger logger = Logger.getLogger(TransactionManager.class);
  /** Database access provider. */
  private ObjectStorage objectStorageProvider = null;
  /** Read only helpers. */
  private Map<String, ISelect> readOnlyHelpers = new HashMap<String, ISelect>();
  /** Read write helpers. */
  private Map<String, ITransact> readWriteHelpers = new HashMap<String, ITransact>();
  /** Write time stamp. */
  private volatile int writeTimeStamp = 0;

  /** Examines validity of call context and throws exception if invalid.
    *
    * @param ctx call context
    * @throws InvalidContextException if the context is invalid.
    */
  private void examineContext(String ctx)
  throws InvalidContextException {
    if (ctx == null) {
      logger.warn("TransactionManager: ISelect or ITransact call with null context.");
      throw new InvalidContextException("Null context.");
    }
    if (ctx.length() <= 0) {
      logger.warn("TransactionManager: ISelect or ITransact call with empty context.");
      throw new InvalidContextException("Empty context.");
    }
  }
  
  /** Increments write time stamp.
    */
  private void incrementTimeStamp() {
    synchronized (this) {
      ++writeTimeStamp; 
    }
  }
  /** Returns current time stamp value.
    * 
    * @return current time stamp value
    */
  private int getCurrentTimeStamp() {
    int timeStamp = 0;
    synchronized (this) {
      timeStamp = writeTimeStamp;
    }
    return timeStamp;
  }

  /** Fetches from cache or creates a select helper (wrapped db session).
    * 
    * @return an implementation of ISelect interface
    * @throws DatabaseUnreachableException
    * @throws InvalidContextException
    */
  private synchronized ISelect getSelectHelper()
  throws DatabaseUnreachableException, InvalidContextException {
    String ctx = SOFAThreadHelper.getCallContext();
    examineContext(ctx);
    ISelect result = readOnlyHelpers.get(ctx);
    if (result == null) {
      if (objectStorageProvider == null) {
        throw new RuntimeException("Required interface must be assigned first.");
      }
      result = new ReadOnlyTransactionHelper(objectStorageProvider);
      readOnlyHelpers.put(ctx, result);
    }
    // update time stamp
    if (result instanceof TransactionHelperBase) {
      ((TransactionHelperBase)result).updateTimeStamp(
          getCurrentTimeStamp());
    } else {
    	logger.error("Select helper not instance of TransactionHelperBase.");
    }
    
    return result;
  }

  /** Fetches from cache or creates a transaction helper (wrapped db session).
    * 
    * @return an implementation of ITransact interface
    * @throws DatabaseUnreachableException
    * @throws InvalidContextException
    */
  private synchronized ITransact getTransactHelper()
  throws DatabaseUnreachableException, InvalidContextException {
    String ctx = SOFAThreadHelper.getCallContext();
    examineContext(ctx);
    ITransact result = readWriteHelpers.get(ctx);
    if (result == null) {
      if (objectStorageProvider == null) {
        throw new RuntimeException("Required interface must be assigned first.");
      }
      result = new MutexTransactionHelper(objectStorageProvider);
      readWriteHelpers.put(ctx, result);
    }
    return result;
  }

  // ISelect methods

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect#select(IQuery) }
    * method.
    */ 
  public <T extends AbstractBase> Objects<T> select(IQuery q)
  throws DatabaseUnreachableException, InvalidContextException {
    return getSelectHelper().select(q);
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect#select(IQuery, int, int) }
    * method.
    */ 
  public <T extends AbstractBase> Objects<T> select(IQuery q,
                                                    int startIndex,
                                                    int endIndex)
  throws DatabaseUnreachableException, InvalidContextException {
    return getSelectHelper().select(q, startIndex, endIndex);
  }

  // ITransact methods

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#fetch(IQuery) }
    * method.
    */ 
  public <T extends AbstractBase> Objects<T> fetch(IQuery q)
  throws DatabaseUnreachableException, TransactionAbortedException {
    return getTransactHelper().fetch(q);
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#fetch(IQuery, int, int) }
    * method.
    */ 
  public <T extends AbstractBase> Objects<T> fetch(IQuery q,
                                                   int startIndex,
                                                   int endIndex)
  throws DatabaseUnreachableException, TransactionAbortedException {
    return getTransactHelper().fetch(q, startIndex, endIndex);
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#store(T) }
    * method.
    */ 
  public <T extends AbstractBase> T store(T object)
  throws DatabaseUnreachableException, TransactionAbortedException {
    return getTransactHelper().store(object);
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#commit() }
    * method.
    */ 
  public boolean commit()
  throws DatabaseUnreachableException {
    ITransact helper;
    try {
      helper = getTransactHelper();      
    } catch (TransactionAbortedException t) {
      return false;
    }
    boolean result = false;
    result = helper.commit();
    if (result == true) {
    	incrementTimeStamp();
    }
    return result;
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#rollback() }
    * method.
    */ 
  public void rollback() 
  throws DatabaseUnreachableException {
    ITransact helper;
    try {
      helper = getTransactHelper();
    } catch (TransactionAbortedException t) {
      return;
    }
    helper.rollback();
  }

  /** Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact#isAborted() }
    * method.
    */ 
  public boolean isAborted() {
    ITransact helper;
    try {
      helper = getTransactHelper();
    } catch (TransactionAbortedException t) {
      return true;
    } catch (DatabaseUnreachableException d) {
      return true;
    }
    return helper.isAborted();
  }

  // SOFAServer method

	public Object getProvided(String name) {
    if (name != null) {
      if (name.equals("transact")) {
        return this;
      } else if (name.equals("select")) {
        return this;
      }
    }
    throw new RuntimeException("Illegal or unknown provided interface name.");
  }

  // SOFAClient method

	public void setRequired(String name, Object iface) {
    if (name != null && name.equals("ostorage")) {
      if (iface instanceof ObjectStorage ) {
        objectStorageProvider = (ObjectStorage) iface;
        return;
      } else { 
        throw new RuntimeException("Interface type mismatch for interface named '" + name + "'.");
      }
    } else {
      throw new RuntimeException("Null or invalid required interface name.");
    }
  }
}
