/* $Id: DataExchangeClient.java 813 2009-11-17 12:05:05Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.de;

import java.util.LinkedList;
import java.util.Properties;
import java.util.Queue;

import org.apache.log4j.Logger;
import org.neodatis.odb.core.query.criteria.ICriterion;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.SOFALifecycle;
import org.objectweb.dsrg.sofa.SOFAParametrized;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.de.IDataExchange;
import org.objectweb.dsrg.sofa.demos.sofashop.de.IInitExchange;


public class DataExchangeClient
		implements IInitExchange, DataExchangeActionContainer,
		           SOFAClient, SOFALifecycle, SOFAParametrized {
	private static Logger logger = Logger.getLogger(DataExchangeClient.class);
	
	private static final String PASSPHRASE_PROPERTY_NAME = "DataExchangeClient.Passphrase";
	private static final String STOREID_PROPERTY_NAME = "DataExchangeClient.StoreID";	
	private static final String REQUIRED_ISELECT_INTERFACE_NAME = "select";
	private static final String REQUIRED_ITRANSACT_INTERFACE_NAME = "update";
	private static final String REQUIRED_IDATAEXCHANGE_INTERFACE_NAME = "exchange";	
	private static final int MAX_ACTION_QUE_SIZE = 3;
	
	/** Passphrase for authentication to the server. */
	private String passphrase = "";
	private long storeId = 0;
	private long sessionId = 0;
	private ITransact dtx = null;
	private ISelect ds = null;
	private IDataExchange de = null;
	private Queue<DataExchangeAction> downloadActionList = new LinkedList<DataExchangeAction>();
	private Queue<DataExchangeAction> uploadActionList = new LinkedList<DataExchangeAction>();
	// request manage - filters too often requests - this might be handled within a microcomponent
	private RequestManager requestManager = new RequestManager();
	
	private Queue<DataExchangeAction> selectQue(int direction) {
		Queue<DataExchangeAction> list = null;
		if (direction == IInitExchange.DIRECTION_GET) {
			list = downloadActionList;
		} else if (direction == IInitExchange.DIRECTION_PUT) {
			list = uploadActionList;
		}
		return list;
	}
	private DataExchangeAction createAction(int direction) {
		logger.trace("Creating new action.");
		if (direction == IInitExchange.DIRECTION_GET) {
			return new DataDownloadAction(dtx, de);
		} else if (direction == IInitExchange.DIRECTION_PUT) {
			return new DataUploadAction(ds, de);
		}
		return null;
	}
	private DataExchangeAction getAction(int direction) {
		DataExchangeAction action = null;
		logger.trace("Call.");
		synchronized (this) {
			Queue<DataExchangeAction> que = selectQue(direction);
			if (que != null) {
				action = que.poll();
				if (action == null) {
					action = createAction(direction);
				}
				action.setSessionId(sessionId);
				
			}
		}
		logger.trace("Return.");
		return action;
	}
	private void returnAction(DataExchangeAction action) {
		int direction = -1;
		synchronized (this) {
			if (action instanceof DataDownloadAction) {
				direction = IInitExchange.DIRECTION_GET;
			} else if (action instanceof DataUploadAction) {
				direction = IInitExchange.DIRECTION_PUT;			
			}
			Queue<DataExchangeAction> que = selectQue(direction);
			if (que != null && que.size() < MAX_ACTION_QUE_SIZE) {
				que.add(action);
			}
		}
	}

	/** @see DataExchangeActionContainer#putAction(DataExchangeAction)
	  * @see #returnAction(DataExchangeAction)
	  */
	@Override
	public void putAction(DataExchangeAction action) {
		returnAction(action);
	}
	/** @see IInitExchange#initExchange(Class, ICriterion, int, boolean)
	  */
	@Override
	public void initExchange(Class<? extends AbstractBase> clazz,
			ICriterion criterion, int direction, boolean block)
			throws DatabaseUnreachableException {
		// ask request manager whether to actually do the data exchange
		if (!requestManager.exchangeConfirmed(clazz, criterion, direction)) {
			// refused -> ignore silently
			return;
		}
		
		try {
			logger.trace("Call.");
			if (!authenticated()) {
				throw new DatabaseUnreachableException(
					"Not authenticated.");
			}
			logger.trace("Get action.");
			DataExchangeAction action = getAction(direction);
			logger.trace("Action obtained.");
			if (block) {
				logger.trace("Invoking from current thread - block.");
				action.perform(clazz, criterion);
				returnAction(action);
				logger.trace("Invokation from current thread finished.");			
			} else {
				logger.trace("Invoking from another thread - immediately return.");			
				Thread t = new Thread(
						new DataExchangeWorker(this, action, clazz, criterion));
				t.start();
			}
		} catch (Throwable t) {
			throw new DatabaseUnreachableException("Internal error: " + t.getClass().getName() + ": " + t.getMessage(), t);
		}
	}
	
	/** @see SOFAClient#setRequired(String, Object)
	  */
	@Override
	public void setRequired(String arg0, Object arg1) {
		if (arg0 == null || arg1 == null) {
			logger.error("Set null required interface.");
			return;
		}
		if (REQUIRED_ISELECT_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof ISelect) {
				ds = (ISelect) arg1;					
			} else {
				logger.error("Incompatibe endpoint for interface name" +
						REQUIRED_ISELECT_INTERFACE_NAME);
			}
		} else if (REQUIRED_ITRANSACT_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof ITransact) {
				dtx = (ITransact) arg1;
			} else {
				logger.error("Incompatibe endpoint for interface name" +
						REQUIRED_ITRANSACT_INTERFACE_NAME);
			}			
		} else if (REQUIRED_IDATAEXCHANGE_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof IDataExchange) {
				de = (IDataExchange) arg1;
			} else {
				logger.error("Incompatibe endpoint for interface name" +
						REQUIRED_IDATAEXCHANGE_INTERFACE_NAME);
			}						
		} else {
			logger.trace("Unkown interface name '" + arg0 + "'.");
		}
	}
	/** Authenticates itself to the server.
	  * 
	  * @see org.objectweb.dsrg.sofa.SOFARunnable#start()
	  */
	@Override
	public void start() {
		logger.trace("start");
	}
	/**
	  * @see org.objectweb.dsrg.sofa.SOFAStoppable#stop()
	  */
	@Override
	public void stop() {
		logger.trace("stop");
	}
	/**
	  * @see org.objectweb.dsrg.sofa.SOFAStoppable#stopping()
	  */
	@Override
	public void stopping() {
		logger.trace("stopping");
	}
	
	/** 
	  * 
	  * @param arg0 properties
	  * @see org.objectweb.dsrg.sofa.SOFAParametrized#setProperties(Properties)
	  */
	@Override
	public void setProperties(Properties arg0) {
		String pp = arg0.getProperty(PASSPHRASE_PROPERTY_NAME);
		if (pp == null) {
			logger.info("Invalid configuration - passphrase not specified. Using empty passphrase.");
		} else {
			synchronized (this) {
				passphrase = pp;
			}
		}
		String sid = arg0.getProperty(STOREID_PROPERTY_NAME);
		if (sid == null) {
			logger.info("Invalid configuration - store id not specified. Using default one.");
		} else {
			long id = 0;
			try {
				id = Long.valueOf(sid).longValue();
				synchronized (this) {
					storeId = id;
				}				
			} catch  (Throwable t) {
				logger.info("Invalid configuration - bad number for store id. Using default one.");
			}
		}
	}
		
	public boolean authenticated() {
		boolean result = false;
		synchronized (this) {
			if (sessionId == 0) {
				try {
					sessionId = AuthenticationHelper.authenticate(
							de, ds, storeId, passphrase);
				} catch (DatabaseUnreachableException due) {
				}
			}
			result = sessionId != 0;
		}
		return result;
	}
}
