/* $Id: DataExchangeClientObjectUpdater.java 805 2009-11-09 02:20:23Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.de;

import java.util.Collection;

import org.apache.log4j.Logger;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.ICriterion;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.util.neodatisodb.AbstractNeodatisODBObjectUpdater;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;

/**
 * @author Ondrej Cerny
 *
 */
public class DataExchangeClientObjectUpdater extends AbstractNeodatisODBObjectUpdater {
	private static Logger logger = Logger.getLogger(DataExchangeClientObjectUpdater.class);
	
	public DataExchangeClientObjectUpdater(ITransact tx) {
		this.tx = tx;
	}
	
	/** Transaction provider. */
	private ITransact tx = null;
	
	/** Attempts once to execute update transaction.
	  * 
	  * @param o object to update a local version of it
	  * @throws DatabaseUnreachableException when database is not reachable
	  * @throws TransactionAbortedException when the transacation aborted
	  * @throws IllegalArgumentException when the object is not updatable 
	  */
	public void updateObjectInDatabaseByObject(Object o)
	throws DatabaseUnreachableException, TransactionAbortedException, IllegalArgumentException {
		
		logger.trace("Call.");
		
		// ensure new transaction (todo)
		
		AbstractBase result = null;
		boolean commit = true;
		try {
			try {
				if (o instanceof Product) {
					logger.trace("Get updated replica of a Product.");
					result = (AbstractBase)getUpdatedReplica((Product)o);
				} else if (o instanceof Order) {
					logger.trace("Get updated replica of a Order.");					
					result = (AbstractBase)getUpdatedReplica((Order)o);
				} else {
					logger.error("Unable to get updated replica of an unsupported object.");										
					throw new IllegalArgumentException(
							"Not updatable.");
				}
			} catch (IllegalStateException ise) {
				Throwable t = ise.getCause();
				if (t != null && t instanceof TransactionAbortedException) {
					throw (TransactionAbortedException)t;
				} else if (t != null && t instanceof DatabaseUnreachableException) {
						throw (DatabaseUnreachableException)t;
				} else {
					throw ise;
				}
			}
			tx.store((AbstractBase) result);
		} catch (TransactionAbortedException tae) {
			commit = true;
		} catch (Throwable t) {
			commit = false;
		} finally {
			if (commit) {
				if (tx.commit()) {
					logger.trace("Update commited.");
					return;
				} else {
					logger.trace("Update commit failed.");					
					throw new TransactionAbortedException(
							"Aborted on commit.");
				}
			} else {
				logger.trace("Update aborted - rolling back.");									
				tx.rollback();
			}
		}		
	}

	/**
	 * @see org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.util.neodatisodb.AbstractNeodatisODBObjectUpdater#contentReleased(java.lang.Object)
	 */
	@Override
	protected void contentReleased(Object o) {
		
		//TODO: should remove objects from database to save space
		
		logger.trace("Content released: " + o.getClass().getName() + ": "+ o.toString());
	}
	
	/**
	 * 
	 */
	@Override
	protected <T extends AbstractBase> T objectCreated(T o) {
		try {
			return (T)tx.store(o);			
		} catch (TransactionAbortedException tae) {
			throw new IllegalStateException(
					"Transaction aborted.", tae);
		} catch (DatabaseUnreachableException due) {
			throw new IllegalStateException(
					"Database unreachable.", due);			
		}		
	}

	/**
	 * @see org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.util.neodatisodb.AbstractNeodatisODBObjectUpdater#getReplica(java.lang.Class, org.neodatis.odb.core.query.criteria.ICriterion, java.lang.String)
	 */
	@Override
	protected Object getReplica(Class<?> clazz, ICriterion idCrit,
			String storeReferenceFieldName) {
		logger.trace("Getting replica of " + clazz.getName() + " where " + idCrit.toString() + ".");
		IQuery q = new CriteriaQuery(clazz, idCrit);
		Object result = null;
		Collection<AbstractBase> results = null;		
		try {
			results = tx.fetch(q);
		} catch (TransactionAbortedException tae) {
			throw new IllegalStateException(
					"Transaction aborted.", tae);
		} catch (DatabaseUnreachableException due) {
			throw new IllegalStateException(
					"Database unreachable.", due);			
		}
		if (results != null) {
			if (results.size() == 1) {
				logger.trace("Have matching object!");
				for (Object o: results) {
					result = o;
					break;
				}
			} else if (results.size() > 1) {
				logger.warn("More than one matching objects! Returning null!");
			}
		} else {
			logger.warn("Null objects! Returning null!");
		}
		return result;
	}

}
