/* $Id: DataExchangeDownloadHelper.java 766 2009-10-17 23:52:31Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.de;

import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.neodatis.odb.Objects;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.ICriterion;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;
import org.neodatis.odb.impl.core.query.list.objects.SimpleList;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Store;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.SupplyTerms;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException;

/**
 * @author Ondrej Cerny
 *
 */
public class DataExchangeDownloadHelper {
	/** Logger. */
	private static Logger logger = Logger.getLogger(DataExchangeDownloadHelper.class);
	
	
	/** Gets objects with given characteristics from database.
	  * 
	  * @param <T>
	  * @param database
	  * @param s
	  * @param clazz
	  * @param criteria
	  * @return
	  * @throws DatabaseUnreachableException
	  * @throws DataExchangeRefusedException
	  */
	@SuppressWarnings("unchecked")
	public static <T extends AbstractBase> Objects<T> downloadObjects(ISelect database, 
			Store s, Class<T> clazz, ICriterion criteria)
			throws DatabaseUnreachableException, DataExchangeRefusedException {
		if (database == null) {
			logger.trace("Bad parameter: database is null.");
			throw new DatabaseUnreachableException(
					"Database provider not initialized properly.");
		}
		if (clazz == null) {
			logger.trace("Bad parameter: clazz is null.");
			throw new DataExchangeRefusedException(
					"Unable to provide objects of null class.");
		}
		logger.trace("call - clazz: " + clazz.getName() +
				//criteria != null ? " criteria: " + criteria.toString() : "" +
				".");
		if (Product.class.equals(clazz)) {
			return (Objects<T>)downloadProducts(database, s, Product.class, criteria);
		} else if (Order.class.equals(clazz)) {
			return (Objects<T>)downloadOrders(database, s, Order.class, criteria);
		} 
		logger.trace("clazz did not match");
		throw new DataExchangeRefusedException(
				"Incompatible type requested for download.");
	}
	
	
	/** Get products with just store-related supply terms.
	  * 
	  * @param database database
	  * @param s store
	  * @param clazz product class
	  * @param criteria additional criteria 
	  * @return collection of products
	  * @throws DatabaseUnreachableException
	  * @throws DataExchangeRefusedException
	  */
	private static Objects<Product> downloadProducts(ISelect database, 
			Store s, Class<Product> clazz, ICriterion criteria)
			throws DatabaseUnreachableException, DataExchangeRefusedException {
		
		if (s == null) {
			logger.trace("Null store.");
			throw new DataExchangeRefusedException(
					"Unable to download objects for null store.");
		}
		long storeId = s.getId();
		logger.trace("downloading products for store with id " + storeId);
		
		// get products from database
		logger.trace("About to get products from database.");
		Objects<Product> products = getObjects(database, clazz, criteria);
		
		Objects<Product> filtered = new SimpleList<Product>();
		
		// return if empty
		if (products == null || products.size() < 1) {
			logger.trace("Returning empty collection of products.");
			return filtered;
		}
		
		// filter out products that do not have any supply terms for the given store
		logger.trace("Filter out not store-related products.");
		for (Product p : products) {
			// filter out unusable supply terms
			if (p.getSupplyTerms() != null) {
				boolean matches = false;
				List<SupplyTerms> originalList = p.getSupplyTerms();
				List<SupplyTerms> tmpList = new Vector<SupplyTerms>(originalList);
				originalList.clear();
				for (SupplyTerms st : tmpList) {
					Store store = st != null ? st.getStore() : null;
					if (store != null && store.getId() == storeId) {
						matches = true;
						originalList.add(st);
					}
				}
				if (matches) {
					p.setSupplyTerms(originalList);
					filtered.add(p);
				}
			}
		}
		logger.trace("Returning filtered collection of " +
				filtered.size() + " products.");
		return filtered;
	}
	
	private static Objects<Order> downloadOrders(ISelect database, 
			Store s, Class<Order> clazz, ICriterion criteria)
			throws DatabaseUnreachableException, DataExchangeRefusedException {
		throw new DataExchangeRefusedException(
				"Order download not implemented yet.");
		
	}
	
	/** Returns collection of objects from database.
	  * 
	  * 
	  * @param database database select provider
	  * @param clazz objects class
	  * @param criteria criteria
	  * @return collection of objects
	  * @throws DataExchangeRefusedException
	  * @throws DatabaseUnreachableException
	  */
	private static <T extends AbstractBase> Objects<T> getObjects(ISelect database,
			Class<T> clazz, ICriterion criteria)
			throws DataExchangeRefusedException, DatabaseUnreachableException {
		// database query
		IQuery  q = null;
		if (criteria == null) {
			// no criteria - all objects of given type
			q = new CriteriaQuery(clazz);
		} else {
			// criteria given - pre-filter the objects
			q = new CriteriaQuery(clazz, criteria);
		}
			
		// get objects from database
		Objects<T> objs = null;
		try {
			objs = database.select(q);
		} catch (InvalidContextException ice) {
			logger.warn("Invalid context exception occured.");
			throw new DatabaseUnreachableException(
					"Error conecting database from data exchange server.", ice);
		} catch (DatabaseUnreachableException due) {
			logger.trace("Database unreachable caught.");
			throw due;
		} catch (Throwable t) {
			logger.error("Unkown exception caught on " +
					"data exchange server (download) " +
					"when getting data from database. (" +
					t.getClass().getName() + ": " +
					t.getMessage());
			throw new DatabaseUnreachableException(
					"Unkown database error.", t);
		}
		
		// log result info
		logger.trace("Object collection fetched from database: " +
				objs == null ? "null" : "size=" + objs.size());
		return objs;
	}
}
