/* $Id: DataExchangeServer.java 766 2009-10-17 23:52:31Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.de;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.neodatis.odb.Objects;
import org.neodatis.odb.core.query.criteria.ICriterion;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.AbstractBase;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Store;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;

/**
 * @author Ondrej Cerny
 *
 */
public class DataExchangeServer implements SOFAClient, IDataExchange {
	private static final String REQUIRED_ISELECT_INTERFACE_NAME = "select";
	private static final String REQUIRED_ITRANSACT_INTERFACE_NAME = "update"; 	
	/** Logger. */
	private static Logger logger = Logger.getLogger(DataExchangeServer.class);
	/** Session map. */
	private Map<Long, Store> sessionMap = new HashMap<Long, Store>();
	/** Last released session ID. */
	private long lastUsedSessionId = 0;
	/** Has the session ID counter already overflow?. */
	private boolean nextRun = false;
	/** ISelect provider. */
	private ISelect dselect = null;
	/** ITransact provider. */
	private ITransact dtx = null;
	
	/** Returns an available session id (including 0).
	  * 
	  * @return session id
	  */
	private synchronized long getUncheckedSessionId() {
			long previousLasUsedSessionId = lastUsedSessionId;
			++lastUsedSessionId;
			if (previousLasUsedSessionId > lastUsedSessionId) {
				nextRun = true;
			}
			while (nextRun) {
				if (!sessionMap.containsKey(new Long(lastUsedSessionId))) {
					break;
				}
				++lastUsedSessionId;
			}
			return lastUsedSessionId;
	}
	
	/** Returns an available non-zero session ID.
	  * 
	  * @return non-zero session id
	  */
	private long getAvailableSessionId() {
		long id = getUncheckedSessionId();
		if (id == 0) {
			id = getUncheckedSessionId();
		}
		if (id == 0) {
			throw new IllegalStateException(
					"Impossible happend: all session id's are occupied!");
		}
		return id;
	}
	
	/** Returns {@link Store store} object for given session ID.
	  * 
	  * @param sessionId
	  * @return {@link Store store} object
	  *         or null if the session ID is not valid
	  */
	private Store getStoreForSessionId(long sessionId) {
		Store result = null;
		synchronized (sessionMap) {
			result = sessionMap.get(new Long(sessionId));
		}
		return result;
	}
	

	/**
	 * @see org.objectweb.dsrg.sofa.demos.sofashop.de.IDataExchange#authenticate(org.objectweb.dsrg.sofa.demos.sofashop.de.Credentials)
	 */
	@Override
	public long authenticate(Credentials credentials) {
		long id = 0;
		if (credentials != null) {
			Store s = credentials.getStore();
			if (s != null) {
				id = getAvailableSessionId();
				
				// TODO : check password (on success, id is non-zero)
				
				if (id != 0) {
					synchronized (sessionMap) {
						sessionMap.put(new Long(id), s);
					}
					logger.info("Authentication of store " + s.getId() + " succeed. Session ID " + id + ".");
				} else {
					logger.info("Authentication of store " + s.getId() + " failed. Not authorized.");
				}
			} else {
				logger.info("Unable to associate session with null store.");
			}
		}
		return id;
	}

	/**
	 * @see org.objectweb.dsrg.sofa.demos.sofashop.de.IDataExchange#downloadObjects(long, java.lang.Class, org.neodatis.odb.core.query.criteria.ICriterion)
	 * @see DataExchangeDownloadHelper
	 */
	@Override
	public <T extends AbstractBase> Objects<T> downloadObjects(long sessionId,
			Class<T> clazz, ICriterion criteria)
			throws DatabaseUnreachableException, DataExchangeRefusedException {
		Store store = getStoreForSessionId(sessionId);
		if (store != null) {
			return DataExchangeDownloadHelper.downloadObjects(dselect, store, clazz, criteria);			
		} else {
			throw new DataExchangeRefusedException(
					"Session not found.");
		}
	}
	
	/**
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.de.IDataExchange#uploadObjects(java.util.List)
	  * @see DataExchangeUploadHelper
	  */
	@Override
	public <O extends AbstractBase> void uploadObjects(long sessionId, List<O> ojects)
		throws DatabaseUnreachableException,
		       DataExchangeRefusedException {
		Store store = getStoreForSessionId(sessionId);
		if (store != null) {
			DataExchangeUploadHelper.uploadObjects(dtx, store, ojects);		
		} else {
			throw new DataExchangeRefusedException(
					"Session not found.");
		}
		
	}

	


	/**
	 * @see org.objectweb.dsrg.sofa.SOFAClient#setRequired(java.lang.String, java.lang.Object)
	 */
	@Override
	public void setRequired(String arg0, Object arg1) {
		if (arg1 == null) {
			logger.error("setRequired: null provider");
			return;
		}
		if (arg0 == null) {
			logger.error("setRequired: null interface name");
			return;
		}
		if (REQUIRED_ISELECT_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof ISelect) {
				dselect = (ISelect) arg1;
			} else {
				logger.error("Interface " + REQUIRED_ISELECT_INTERFACE_NAME +
						" provider is not instance of ISelect.");
				
			}
		} else if (REQUIRED_ITRANSACT_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof ITransact) {
				dtx = (ITransact) arg1;
			} else {
				logger.error("Interface " + REQUIRED_ITRANSACT_INTERFACE_NAME +
						" provider is not instance of ISelect.");
				
			}
		} else {
			logger.error("Unkown interface name " + arg0 + ".");
		}

	}

}
