/* $Id: Order.java 762 2009-10-12 11:19:00Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.inv.order;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.SOFAServer;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;

/**
  * 
  * @author Ondrej Cerny
  *
  */
public class Order implements IOrder, SOFAServer, SOFAClient {
	
	/** Name of transaction interface. */
	private static final String DB_TRANSACT_REQUIRED_INERFACE_NAME =
		"data";
	/** Name of order interface. */
	private static final String ORDER_PROVIDED_INTERFACE_NAME =
		"order";
	/** Maximal number of unsuccessfull transaction attempts. */
	private static final int MAX_TX_ATTEMPTS = 100;
	/** Logger. */
	private static Logger logger = Logger.getLogger(Order.class);
	/** Transaction provider. */
	private ITransact tx = null;
	/** Order registration action. */
	private OrderRegistationAction registrationAction;
	/** Order receive action. */
	private OrderCloseAction receiveAction;
	
	/** Handles i-th error of transaction.
	  * 
	  * @param i number of previous errors
	  * @param e error exception
	  * @return number of errors including this one
	  * @throws InventoryUnreachableException when it is considered as failure
	  */
	private static int handleTxError(int i, Exception e) 
	throws InventoryUnreachableException {
		if (i >= MAX_TX_ATTEMPTS) {
			throw new InventoryUnreachableException(
					"Unable to process the transaction.",
					e);
		}
		return i + 1;
	}
	
	/** Returns initialized {@link OrderRegistationAction}.
	  *
	  * @return initialized {@link OrderRegistationAction}
	  * @throws InventoryUnreachableException when this component
	  *         has not been initialized properly 
	  */
	protected OrderRegistationAction getOrderRegistrationAction()
			throws InventoryUnreachableException {
		if (registrationAction == null) {
			if (tx == null) {
				throw new InventoryUnreachableException(
						"Unable to initialize order registry.");
			}
			registrationAction = new OrderRegistationAction(tx);
		}
		return registrationAction;
	}
	
	/** Returns initialized {@link OrderCloseAction}.
	  * 
	  * @return initialized {@link OrderCloseAction}
	  * @throws InventoryUnreachableException when this component
	  *         has not been initialized properly
	 */
	protected OrderCloseAction getOrderCloseAction()
			throws InventoryUnreachableException {
		if (receiveAction == null) {
			if (tx == null) {
				throw new InventoryUnreachableException(
				"Unable to initialize order receiver.");
			}
			receiveAction = new OrderCloseAction(tx);
		}
		return receiveAction;
	}
	
	/** Performs given action on given order.
	  * 
	  * @param action
	  * @param order
	  * @throws InventoryUnreachableException on lower level failure
	  * @throws OrderInconsistentException on higher level failure
	  */
	private static org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order performAction(
			OrderAction action,
			org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order order)
			throws InventoryUnreachableException, OrderInconsistentException {
		int attempts = 0;
		org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order result = null;
		try {
			result = action.perform(order);
		} catch (TransactionAbortedException tae) {
			attempts = handleTxError(attempts, tae);
		} catch (DatabaseUnreachableException due) {
			attempts = handleTxError(attempts, due);
		}
		return result;
	}
	                                        
	/** Marks given
	  * {@link org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order order}
	  * as received (goods delivered).
	  * 
	  * <p>
	  * Uses {@link org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order#setCloseDate(java.util.Date)}
	  * for setting current time which also indicates, that the order has
	  * been received.
	  * </p>
	  * 
	  * @see IOrder#receiveOrder(org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order)
	  */
	@Override
	public void receiveOrder(
			String orderCode)
			throws InventoryUnreachableException, OrderInconsistentException {
		org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order dummyOrder =
			new org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order();
		dummyOrder.setCode(orderCode);
		performAction(getOrderCloseAction(), dummyOrder);
	}
	
	/** Creates order in database.
	  * 
	  * <p>
	  * Implements
	  * {@link IOrder#registerOrder(org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order) }.
	  * </p>
	  * <p>
	  * Calls 
	  * {@link OrderRegistationAction#perform(org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order)
	  * </p>
	  * 
	  * @see IOrder
	  * @see OrderRegistationAction
	  */
	@Override
	public String registerOrder(
			org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order o)
			throws InventoryUnreachableException, OrderInconsistentException {
		try {
			return performAction(getOrderRegistrationAction(), o).getCode();
		} catch (NullPointerException npe) {
			return null;
		}
	}

	// SOFAServer method:
	/** Returns this.
	  * 
	  * @see SOFAServer#getProvided(String) 
	  */
	@Override
	public Object getProvided(String arg0) {
		if (ORDER_PROVIDED_INTERFACE_NAME.equals(arg0)) {
				return this;
		}
		logger.error("Request for unknown interface.");			
		return null;
	}	
	// SOFAClient method:
	/** Sets required interface endpoint.
	  * 
	  * @see SOFAClient#setRequired(String, Object) 
	  */
	@Override
	public void setRequired(String arg0, Object arg1) {
		if (DB_TRANSACT_REQUIRED_INERFACE_NAME.equals(arg0)) {
			if (arg1 == null) {
				logger.error(
						"Null '" + arg0 + "' interface endpoint received.");
				throw new NullPointerException(
						"Null interface endpoint received.");
			} else {
				if (arg1 instanceof ITransact) {
					tx = (ITransact)arg1;
				} else {
					logger.error("Incompatible endpoint for interface '" +
							arg0 + "'.");
					throw new ClassCastException(
							"Incompatible conn. endpoint type. ITransact expected.");
				}
			}
		} else {
			logger.error("Unkown interface received.");
		}
	}
}
