/* $Id: OrderCloseAction.java 762 2009-10-12 11:19:00Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.inv.order;

import java.util.Collection;

import org.apache.log4j.Logger;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.OrderItem;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.StockItem;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;

/**
  * 
  * @author Ondrej Cerny
  *
  */
public class OrderCloseAction implements OrderAction {
	
	/** Logger. */
	private Logger logger = Logger.getLogger(OrderCloseAction.class);
	 
	/** Transaction provider. */
	private ITransact tx; 

	
	/** Creates this action bound to transaction provider.
	  * 
	  * @param tx
	  */
	public OrderCloseAction(ITransact tx) {
		this.tx = tx;
	}
	
	/** Updates or creates stock item for each order item.
	  * 
	  * @param o
	  * @throws TransactionAbortedException
	  * @throws DatabaseUnreachableException
	  * @throws OrderInconsistentException
	  */
	private void updateStockItems(Order o)
	throws TransactionAbortedException,
	       DatabaseUnreachableException,
	       OrderInconsistentException {
		if (o != null && o.getOrderItems() != null) {
			for (OrderItem i : o.getOrderItems()) {
				Product p = i != null ? i.getProduct() : null;
				String ean = p != null ? p.getDefaultEAN() : null;
				
				// get stock item with same EAN
				StockItem si = null;
				if (ean != null) {
					Collection<StockItem> stockItems = 
						tx.fetch(new CriteriaQuery(StockItem.class, Where.equal("EAN", ean)));
					if (stockItems != null && stockItems.size() == 1) {
						// get first object
						si = stockItems.iterator().next();
						if (si == null) {
							logger.warn("Watch out! First object null!");
						}
					} else if (stockItems != null && stockItems.size() > 1) {
						String message = "Inconsistent database state: more than 1 StockItem with same EAN " + ean + ".";
						logger.error(message);
						throw new TransactionAbortedException(message);
					}
					if (si == null) {
						logger.trace("Creating new stock item with EAN " + ean + ".");
						si = new StockItem();
						si.setAmount(i.getAmount());
						si.setEAN(ean);
						si.setPrice(p.getDefaultPrice());
						si.setProduct(p);
						si.setStore(o.getStore());
					} else {
						logger.trace("Updating existing stock item with EAN " + ean + ".");
						si.setAmount( si.getAmount() + i.getAmount());
					}
					
					// save changes
					tx.store(si);
					
				} else {
					logger.error(
							"Unable to close inconsisten order " +
							"(product default EAN is null).");
					throw new OrderInconsistentException(
							"Default EAN of a product is null.");
				}
			}
		}
	}
	
	/** Performs this action.
	  *
	  * <p>
	  * Marks {@link Order} as delivered or throws
	  * {@link OrderInconsistentException exception}
	  * when it has been marked already or is somehow
	  * inconsistent.
	  * </p>
	  * 
	  * @param o order with same code as the order to be closed
	  * @throws OrderInconsistentException when the {@link Order order}
	  *         has been already delivered or is somehow inconsistent
	  * @throws DatabaseUnreachableException when the db is not reachable
	  * @throws TransactionAbortedException when the transaction has been
	  */
	public Order perform(Order o)
			throws OrderInconsistentException,
			       DatabaseUnreachableException,
			       TransactionAbortedException {
		Order result = null;
		try {
			String code = o != null ? o.getCode() : null;
			if (code == null) {
				throw new OrderInconsistentException("Null order or its code.");
			}			
			IQuery q = new CriteriaQuery(Order.class,
							Where.equal("code", code));
			Collection<Order> orders = tx.fetch(q);
			if (orders == null || orders.size() != 1) {
				throw new OrderInconsistentException(
						"Unable to identify order in database uniquely.");
			}
			for (Order updatedO: orders) {
				// step 1: update stock items
				updateStockItems(updatedO);
				
				// step 2: update order entry
				
				java.util.Date date;
				
				// check not cancelled
				date = updatedO.getCancelDate();
				if (date != null) {
					throw new OrderInconsistentException(
							"Order has been cancelled.");
				}
				
				// update close date
				date = updatedO.getCloseDate();
				if (date != null) {
					throw new OrderInconsistentException(
							"Order has been already closed.");
				}
				updatedO.setCloseDate(new java.util.Date());
				
				// save changes
				result = tx.store(updatedO);
				
				break;
			}
		} catch (OrderInconsistentException oie) {
			logger.info("Order closing failed: " + oie.getMessage());
			throw oie;
		} finally {
			if (!tx.commit()) {
				String msg = "Transaction aborted on commit.";
				logger.trace(msg);
				throw new TransactionAbortedException(msg);
			}
		}	
		
		return result;
	}
}
