/* $Id: OrderRegistationAction.java 763 2009-10-16 23:16:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.inv.order;

import java.util.Collection;
import java.util.List;

import org.apache.log4j.Logger;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.OrderItem;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Supplier;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.SupplyTerms;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;

/**
  * @author Ondrej Cerny
  *
  */
public class OrderRegistationAction implements OrderAction {
	/** Logger. */
	private static Logger logger = Logger.getLogger(OrderRegistationAction.class);
	/** Transaction provider. */
	ITransact tx;
	
	/** Creates new action bound to transaction provider.
	  * 
	  */
	public OrderRegistationAction(ITransact tx) {
		this.tx = tx;
	}

	/** Returns true if and only if given
	  * {@link SupplyTerms supply terms}
	  * are equivalent. 
	  * 
	  * @param st1 {@link SupplyTerms}
	  * @param st2 {@link SupplyTerms}
	  * @return boolean
	  */
	private boolean isEquivalentSupplyTerms(SupplyTerms st1, SupplyTerms st2) {
		if (st1 == st2) {
			return true;
		}
		if (st1 == null || st2 == null) {
			return false;
		}
		return 
				st1.getLowerLimitIncl() == st2.getLowerLimitIncl()
			&&	st1.getUpperLimitExcl() == st2.getUpperLimitExcl()
			&&	st1.getPrice() == st2.getPrice()
			&&  st1.getValidFrom().equals(st2.getValidFrom())
			&&  st1.getValidTo().equals(st2.getValidTo())
			&&  st1.getSupplier().getRegNo().equals(st2.getSupplier().getRegNo());
	}
	
	/** Validates {@link SupplyTerms supply terms}.
	  * 
	  * @param st {@link SupplyTerms} for validation
	  * @param s {@link Supplier supplier} for matching
	  * @param oi {@link OrderItem order item} (for matching amount)
	  * @throws OrderInconsistentException when the supply terms are invalid
	  */
	private void validateSupplyTerms(
			SupplyTerms st,
			Supplier s,
			OrderItem oi)
		throws OrderInconsistentException {
		try {
			if (!s.getRegNo().equals(st.getSupplier().getRegNo())) {
				throw new OrderInconsistentException(
						"Mismatching supplier.");
			}
			java.util.Date now = new java.util.Date();
			if (now.before(st.getValidFrom())) {
				throw new OrderInconsistentException(
						"Supply terms not valid yet: from = " + st.getValidFrom().toString() + " now = " + now.toString());
			}
			if (now.after(st.getValidTo())) {
				throw new OrderInconsistentException(
						"Supply terms expired: to = " + st.getValidTo().toString() + " now = " + now.toString());
			}
			double amount = oi.getAmount();
			if (st.getLowerLimitIncl() > amount) {
				throw new OrderInconsistentException(
				"Supply temrs not applicable (amount under limit).");					
			}
      double upperLimit = st.getUpperLimitExcl();
			if (upperLimit >= 0 && upperLimit <= amount) {
				throw new OrderInconsistentException(
				"Supply temrs not applicable (amount over limit).");					
			}				
		} catch (OrderInconsistentException oie) {
			logger.info("Supply terms validation failed: " + oie.getMessage());
			throw oie;
		} catch (Throwable t) {
			logger.info("Supply terms validation failed: " + t.getMessage());
			throw new OrderInconsistentException(
					"Supply terms inconsistent.");			
		}
	}
	
	/** Re-binds and validates associations of
	  * {@link org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order order}
	  * and its {@link OrderItem order items}.
	  * 
	  * <p>
	  * Re-binds {@link Supplier supplier} and calls
	  * {@link #rebindAssociationsAndValidate(OrderItem, Supplier)}
	  * for each {@link OrderItem order item}.
	  * </p> 
	  * 
	  * @param o {@link org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order}
	  * @throws OrderInconsistentException when the order or its order items are
	  *         inconsistent (no order items, no or mismatching supplier...)  
	  * @throws TransactionAbortedException
	  * @throws DatabaseUnreachableException
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order#setSupplier(Supplier) 
	  */
	private void rebindAssociationsAndValidate(
			org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order o)
	throws OrderInconsistentException,
	       TransactionAbortedException,
	       DatabaseUnreachableException  {
		Supplier s = o != null ? o.getSupplier() : null;
		if (s != null) {
			List<OrderItem> orderItems = o.getOrderItems();
			
			// validate order items part 1
			if (orderItems == null || orderItems.size() < 1) {
				String msg = "No order items - registering such order is a non-sense.";
				logger.info("Rebinding order associations: " + msg);
				throw new OrderInconsistentException(msg);
			}
			
			// re-bind and validate supplier
			IQuery q = new CriteriaQuery(
					Supplier.class, Where.equal("regNo", s.getRegNo()));
			
			Collection<Supplier> suppls = tx.fetch(q);
			if (suppls != null && suppls.size() == 1) {
				for (Supplier newS : suppls) {
					// re-bind supplier					
					o.setSupplier(newS);
					
					// re-bind order items' references 
					// and validate them with respect to supplier
					for(OrderItem oi : orderItems) {
						rebindAssociationsAndValidate(oi, newS);
					}
					
					break; // for sure
				}
				
			} else {
				String msg = "Inconsistent supplier reference.";
				logger.info("Rebinding order associations: " + msg);
				throw new OrderInconsistentException(msg);
			}
		} else {
			String msg = "Null order.";
			logger.info("Rebinding order associations: " + msg);
			throw new OrderInconsistentException(msg);
		}
	}
	/** Re-binds associations and validates {@link OrderItem order item}.
	  * 
	  * <p>
	  * Re-binds {@link SupplyTerms} and {@link Product} and check
	  * whether it matcces given {@link Supplier}.
	  * </p>
	  * 
	  * @param oi {@link OrderItem}
	  * @param s {@link Supplier}
	  * @throws OrderInconsistentException on any mismatch
	  * @throws TransactionAbortedException 
	  * @throws DatabaseUnreachableException
	  */
	private void rebindAssociationsAndValidate(OrderItem oi, Supplier s)
	throws OrderInconsistentException,
	       TransactionAbortedException,
	       DatabaseUnreachableException  {
		SupplyTerms st = (oi != null) ? oi.getTerms() : null;
		Product p = (oi != null) ? oi.getProduct() : null;
		try {
			// re-bind product
			IQuery q = new CriteriaQuery(Product.class, Where.equal("id", p.getId()));
			Collection<Product> products= tx.fetch(q);
			if (products == null || products.size() != 1) {
				throw new OrderInconsistentException(
						"Order item inconsistent (unable to rebind product).");
			}
			for (Product newP : products) {
				oi.setProduct(newP);
				break;
			}
			
			// re-bind supply terms
			p = oi.getProduct(); // refresh product		
			SupplyTerms newPST = null;
			for(SupplyTerms aPST: p.getSupplyTerms()) {
				if (isEquivalentSupplyTerms(st, aPST)) {
					newPST = aPST;
					break;
				}
			}
			if (newPST == null) {
				throw new OrderInconsistentException(
						"Supply terms probably changed.");
			}
			
			// validate supply terms
			validateSupplyTerms(newPST, s, oi);
			
			oi.setTerms(newPST);
			
		} catch (OrderInconsistentException oie) {
			logger.info("Rebinding order item associations uncovered inconsistency:" + oie.getMessage());
			throw oie;			
		} catch (Throwable t) {
			logger.info("Rebinding order item associations failed:" + t.getMessage());
			throw new OrderInconsistentException(
					"Inconsisten order item.");
		}
		
	}
 	
	/** Assigns unique code to the order.
	  * 
	  * <p>
	  * Note that this depends on particular data model. 
	  * </p>
	  * 
	  */
	private Order assignCode(Order o)
			throws TransactionAbortedException, DatabaseUnreachableException {
		
		// save the object for updating database object id
		Order newObject = tx.store(o);
		
		// the hash code is database object id of object
		// (just after saving)
		String code = Integer.toString(newObject.hashCode());
		
		// set the code
		newObject.setCode(code);
		
		return newObject;
	}

	/** Marks given
	  * {@link org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order order}
	  * as received (goods delivered).
	  * 
	  * <p>
	  * Uses {@link org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order#setCloseDate(java.util.Date)}
	  * for setting current time which also indicates, that the order has
	  * been received.
	  * </p>
	  * 
	  * @see IOrder#receiveOrder(org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order)
	  */
	public Order perform(
			org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order o)
			throws DatabaseUnreachableException, TransactionAbortedException, OrderInconsistentException {
		Order result = null;
		try {
				
				// re-binds product and supply terms associations
				rebindAssociationsAndValidate(o);
				
				// assign code (which is unique id)
				Order newOrder = assignCode(o);
				
				// assign create date
				newOrder.setCreateDate(new java.util.Date());
				
				// save the order object to the database
				result = tx.store(newOrder);
	
		} finally {
				if (!tx.commit()) {
					String msg = "Transaction aborted on commit.";
					logger.trace(msg);
					throw new TransactionAbortedException(msg);
				}
		}
		return result;
	}
}
