/* $Id: ListProducts.java 766 2009-10-17 23:52:31Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.inv.prod;

import org.apache.log4j.Logger;
import org.neodatis.odb.Objects;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.ICriterion;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.TargetUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.de.IInitExchange;

/**
 * @author Ondrej Cerny
 */
public class ListProducts implements IListProducts, SOFAClient {
	private static Logger logger = Logger.getLogger(ListProducts.class);
	
	private static final String REQUIRED_ISELECT_INTERFACE_NAME =  "data";
	private static final String REQUIRED_IINITEXCHANGE_INTERFACE_NAME =  "iexchange";
	
	/** Data provider. */
	private ISelect db;
	/** Exchanges initiator. */
	private IInitExchange ie;
	
	/** Selects products matching criterion.
	  * 
	  * @param crit criterion
	  * @param start start index
	  * @param end end index
	  * @return selected products
	  * @throws InventoryUnreachableException
	  */
	private Objects<Product> selectProducts(ICriterion crit, int start, int end)
	throws InventoryUnreachableException  {
		if (db == null || ie == null) {
			throw new InventoryUnreachableException(
					"Product listing component not initialized properly.");
		}
		// data exchange - refresh data 
		logger.trace("initiate data exchange - download from enterprise database");
		try {
			ie.initExchange(Product.class, null, IInitExchange.DIRECTION_GET, true);
		} catch (DatabaseUnreachableException due) {
			logger.warn("Data exchange failed - " +
					"database unreachable: " + due.getMessage(), due);
		}
		
		logger.trace("select products from local database");			
		// query and select
		IQuery q = new CriteriaQuery(Product.class, crit);
		Objects<Product> products = null;		
		try {
			products = db.select(q, start, end);
		} catch (Throwable t) {
			logger.trace("Error occured on inventory listing: " +
					t.getClass().getName() + ": " +
					t.getMessage());
		}
		return products;
	}

	/** @see IListProducts#getProductById(long)
	  */
	@Override
	public Product getProductById(long productId)
			throws InventoryUnreachableException {
		Objects<Product> res = selectProducts(Where.equal("id", productId),
				0, 2);
		if (res != null) {
			if (res.size() == 1) {
				return res.getFirst();				
			} else if (res.size() < 1) {
				return null;
			} else {
				logger.error("More than one product mathces id " + productId);
				return null;
			}
		} else {
			return null;
		}
	}
	/** @see IListProducts#listProducts(long, String, int, int)
	  */
	@Override
	public Objects<Product> listProducts(long categoryId, String namePattern,
			int startIndex, int endIndex) throws InventoryUnreachableException {
		ICriterion c = Where.equal("category.id", categoryId);
		if (namePattern != null) {
			ICriterion tmp = Where.and()
				.add(Where.like("name", namePattern))
				.add(c);
			c = tmp;
		}
		return selectProducts(c, startIndex, endIndex);
	}
	/** @see IListProducts#listProducts(String, int, int)
	  */
	@Override
	public Objects<Product> listProducts(String namePattern, int startIndex,
			int endIndex) throws InventoryUnreachableException {
		return selectProducts(
				namePattern != null 
						? Where.like("name", namePattern)
						: null,
				startIndex, endIndex);
	}

	/** @see SOFAClient#setRequired(String, Object)
	  */
	@Override
	public void setRequired(String arg0, Object arg1) {
		if (arg0 == null || arg1 == null) {
			logger.error(
					"Null requiered interface name or endpoind received.");
		} else if (REQUIRED_IINITEXCHANGE_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof IInitExchange) {
				ie = (IInitExchange)arg1;
			} else {
				logger.error("Incompatible endpoint for interface " +
						REQUIRED_IINITEXCHANGE_INTERFACE_NAME + " " +
						"received. Ignoring");
			}
		} else if (REQUIRED_ISELECT_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof ISelect) {
				db = (ISelect)arg1;
			} else {
				logger.error("Incompatible endpoint for interface " +
						REQUIRED_ISELECT_INTERFACE_NAME + " " +
						"received. Ignoring");
			}			
		} else {
			logger.error("Unknown interface name '" + arg0 +
					"' received. Ignoring.");
		}
	}
}
