/* $Id: ListStock.java 652 2009-08-24 00:17:16Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.inv.stock;

import org.objectweb.dsrg.sofa.SOFALifecycle;
import org.objectweb.dsrg.sofa.SOFAClient;

import org.apache.log4j.Logger;

import org.neodatis.odb.Objects;
import org.neodatis.odb.impl.core.query.list.objects.SimpleList;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;

import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.StockItem;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;

/** Provides listing of StockItems.
  * 
  * @author Ondrej Cerny
  */
public class ListStock implements IListStock, SOFALifecycle, SOFAClient {

  /** Logger. */
  private static Logger logger = Logger.getLogger(ListStock.class);

  /** Indicates active state of this component. */
  private boolean isActive;

  /** Database representation. */
  private ISelect database;

  /** Number of retries after
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException}
    * .
    */
  private static final int INVALID_CONTEXT_RETRIES = 10;

  /** Constructor. */
  public ListStock() {
    database = null;
    isActive = false;
  }

  /** Checks whether the component is active ant throws
    * InventoryUnreachableException otherwise.
    * 
    * @throws InventoryUnreachableException when this component is not in
    *                                       active state
    */
  private void isActiveOrThrow() throws InventoryUnreachableException {
    if (!isActive) {
      logger.warn("StockItem listing requested but the ListStock component " +
                  "is not in active state.");
      throw new InventoryUnreachableException("ListStock component is " +
                  "going to stop or has not been activated yet.");
    }
    if (database == null) {
      logger.warn("StockItem listing requested but the database interface " + 
                  "has not been established.");
      throw new InventoryUnreachableException("ListStock component has not " +
                  "got required interface 'data' providing database acces.");
    }
  }

  /** Queries the database.
    *
    * @param query database query
    * @param startIndex index of the first returned object
    * @param endIndex index of the first object behind the last returned object
    * @return result of the query
    * @throws InventoryUnreachableException when the database query failed
    *   (bacause of
    *   {@link org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException}
    *   or
    *   {@link org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException}
    *   )
    */
  private Objects<StockItem> query(IQuery query, int startIndex, int endIndex)
  throws InventoryUnreachableException {
    int tries = 0;
    try {
      while (true) {
        try {
          return database.select(query, startIndex, endIndex);
        } catch (InvalidContextException i) {
          logger.warn("ListStock query caught InvalidContextException. Message: " +
            i.getMessage());
          if ((++tries) >= INVALID_CONTEXT_RETRIES) {
            logger.warn("ListStock has permanently invalid call context. Giving up.");
            throw new InventoryUnreachableException("Unable to query " +
              "the database due to permanently invalid call context.", i);
          }
        }
      }
    } catch (DatabaseUnreachableException d) {
      logger.warn("ListStock component could not connect to database.", d);
      throw new InventoryUnreachableException("Can not connect to database.", d);
    }
  }


  // IListStock methods

  /** Returns single StockItem uniquely identified by its EAN.
    *
    * <p>
    * Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.inv.IListStock#getStockItemByEAN(String)}
    * .
    * </p>
    *
    * @param EAN unique identification of StockItem
    * @return StockItem on success, <code>null</code> iff there is no such
    *         StockItem in the database
    * @throws InventoryUnreachableException when the database is unreachable
    *                                       or this component is going to stop
    */
  public StockItem getStockItemByEAN(String EAN)
  throws InventoryUnreachableException {

    // assert the component is in the correct state (throw exception otherwise)
    isActiveOrThrow();

    // query the database
    Objects<StockItem> stockItems = query(
        new CriteriaQuery(StockItem.class, Where.equal("EAN", EAN)),
        0, 1);

    // extract the correct form of the result
    return (stockItems.size() > 0) ? stockItems.getFirst() : null;
  }

  /** Returns StockItems associated with a product which match exactly given
    * id.
    * 
    * <p>
    * Implements
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.inv.IListStock#getStockItemByEAN(String)}
    * .
    * </p>
    *
    * @param id product unique identification
    * @param startIndex index of the first returned object
    * @param endIndex index of the first object behind the last returned object
    * @return A collection of StockItems
    */
  public Objects<StockItem> listStockItems(long id, int startIndex, int endIndex)
  throws InventoryUnreachableException {

    // assert the component is in the correct state (throw exception otherwise)
    isActiveOrThrow();

    // query the database
    Objects<StockItem> stockItems = query(
        new StockItemsByProductIdQuery(id), startIndex, endIndex);
        //TODO: consider ordering by EAN

    return stockItems;
  }

  /** Returns StockItems associated with products with description
    * that match given <code>pattern</code>.
    *
    * <p>
    * Pattern will be matched using java.util.regex with CASE_INSENSITIVE
    * and UNICODE_AWARE partial matching.
    * </p>
    *
    * @param pattern regular expresion
    * @param startIndex index of the first returned object
    * @param endIndex index of the first object behind the last returned object
    * @returns A list of stock items
    */
  public Objects<StockItem> listStockItems(String pattern,
                                           int startIndex, int endIndex)
  throws InventoryUnreachableException {
    StockItemsByProductDescQuery q;

    // assert the component is in the correct state (throw exception otherwise)
    isActiveOrThrow();

    // prepare query
    try {
      q = new StockItemsByProductDescQuery(pattern);
      //TODO: consider ordering by EAN
    } catch (IllegalArgumentException i) {
      logger.warn("ListStock can not create query for matching StockItems" +
                  "associated to a Product with description matching " +
                  "pattern. Returning empty collection of StockItems.", i);
      return new SimpleList<StockItem>(0);
    }

    // execute the query in the database
    Objects<StockItem> stockItems = query(q, startIndex, endIndex);

    return stockItems;
  }

  //SOFALifecycle methods

  /** Initializes this component.
    */
  public void start() {
    isActive = true;
  }

  /** Prepare for stopping this component. */
  public void stopping() {
    isActive = false;
  }

  /** Stop this component. */
  public void stop() {
  }

  // SOFAClient methods

  /** Sets requred interface endpoint.
    *
    * @param name name of the interface
    * @param object implementing instance
    */
  public void setRequired(String name, Object iface) {
    if (name.equals("data")) {
      if (iface instanceof ISelect) {
        database = (ISelect) iface;
      } else {
        throw new RuntimeException("Endpoint of required interface 'data' does not implement interface ISelect.");
      }
    }
    // ignore other interfaces
  } 

}
