/* $Id: SaleRegistry.java 762 2009-10-12 11:19:00Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.inv.stock;

import java.util.Iterator;

import org.objectweb.dsrg.sofa.SOFALifecycle;
import org.objectweb.dsrg.sofa.SOFAClient;

import org.apache.log4j.Logger;

//import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;

import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Sale;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.SaleItem;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.StockItem;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ITransact;
import org.objectweb.dsrg.sofa.demos.sofashop.db.DatabaseUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException;
import org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;

/** Provides listing of StockItems.
  * 
  * @author Ondrej Cerny
  */
public class SaleRegistry implements ISale, SOFALifecycle, SOFAClient {

  /** Logger. */
  private static Logger logger = Logger.getLogger(SaleRegistry.class);

  /** Indicates active state of this component. */
  private volatile boolean isActive;

  /** Database representation. */
  private ITransact database;

  /** Number of retries after
    * {@link org.objectweb.dsrg.sofa.demos.sofashop.db.InvalidContextException}
    * .
    */
  private static final int INVALID_CONTEXT_RETRIES = 10;
  /** Number of retries when a transaction aborts
    * ({@link org.objectweb.dsrg.sofa.demos.sofashop.db.TransactionAbortedException})
    * .
    */
  private static final int TRANSACTION_ABORTED_RETRIES = 10;

  /** Constructor. */
  public SaleRegistry() {
    database = null;
    isActive = false;
  }

  /** Checks whether the component is active and throws
    * InventoryUnreachableException otherwise.
    * 
    * @throws InventoryUnreachableException when this component is not in
    *                                       active state
    */
  private void isActiveOrThrow() throws InventoryUnreachableException {
    if (!isActive) {
      logger.warn("Sale registering requested but the Sale component " +
                  "is not in active state.");
      throw new InventoryUnreachableException("Sale component is " +
                  "going to stop or has not been activated yet.");
    }
    if (database == null) {
      logger.warn("Sale registering requested but the database interface " + 
                  "has not been established ('data' req. interface).");
      throw new InventoryUnreachableException("Sale component has not " +
                  "received endpoint of required interface 'data' yet.");
    }
  }

  /** Updates stock item related to the sale item <code>si</code>.
    *
    * <p>
    * The stock item will be searched by the <code>stockItemEAN</code>
    * filed of the <code>si</code> object and the amount will be updated 
    * and saved.
    * </p>
    *
    * @param si sale item
    */
  private void updateStockItem(SaleItem si)
  throws DatabaseUnreachableException, InvalidContextException,
         TransactionAbortedException {
    
    if (si == null) {
      logger.warn("Sale registering: updateStockItem(null) called. Ignoring.");
      return;
    }

    String siEAN = si.getStockItemEAN();
    
    Iterable<StockItem> sitems = database.fetch(new CriteriaQuery(StockItem.class,
                                                    Where.equal("EAN", siEAN)));
    Iterator<StockItem> sit = (sitems != null) ? sitems.iterator() : null;

    if (sit != null && sit.hasNext()) {
      StockItem sti = sit.next();
      if (sit.hasNext()) {
        logger.warn("Inconsistency found during updating stock item " +
                    "(sale registering): More than one StockItem with EAN " +
                    siEAN + "! Using the 1st.");
      }

      // update the amount field
      sti.setAmount(
        sti.getAmount() - si.getAmount());

      // save the stock item
      database.store(sti);
      logger.trace("Stock item with EAN=\"" + siEAN + "\" updated.");

    } else {
      logger.error("Error during updating stock item (sale registering): " +
                   "No StockItem with EAN " + siEAN + "! " +
                   "Skipping stock item update.");
    }
  }

  /** Re-binds product association.
    * 
    * @param si sale item
    */
  private void rebindProduct(SaleItem si) 
  throws DatabaseUnreachableException, InvalidContextException,
         TransactionAbortedException {

    if (si == null) {
      logger.warn("Sale registering: rebindProduct(null) called. Ignoring.");
      return;
    }

    Product oldProduct = si.getProduct();

    if (oldProduct != null) {
      // do rebind just in this branch
      long productId = oldProduct.getId();

      // find actual Product object in database
      Iterable<Product> prods = database.fetch(new CriteriaQuery(Product.class,
                                                 Where.equal("id", productId)));
      
      Iterator<Product> pit = (prods != null) ? prods.iterator() : null;

      if (pit != null && pit.hasNext()) {
        Product p = pit.next();
        if (pit.hasNext()) {
          logger.warn("Inconsistency found during product rebinding (sale registering): " +
                      "More than one product with " + productId + " found. " +
                      "Using the 1st.");
        }

        // set the product to the sale item
        si.setProduct(p);
        
        logger.trace("Product with id=" + Long.toString(productId) +
        		" rebound.");
      } else {
        logger.error("Product rebindning failed (sale registering): " +
                     "No product with Id " + productId + ". " +
                     "Left original product.");
      }
    } else {
      logger.warn("Registering a sale with a sale itme with " +
                  "null product association.");
    }
  }


  // ISale method:

  /** Registers sale to database.
    *
    * <p>
    * The proper sale registration consists of following steps:<br/>
    * <ul>
    *   <li>
    *     update amount fields of StockItems of products that has been sold
    *     (StockItems with the same value of <code>EAN</code> fields as
    *      SaleItems' <code>stockItemEAN</code> field)
    *   </li>
    *   <li>
    *     rebind Product objects associated to SaleItems
    *   </li>
    *   <li>
    *     save Sale
    *   </li>
    * </ul>
    * All the steps must be provided within single transaction.
    * </p>
    *
    * @param sale object representing single sale
    *             (composition of mulitple sale items related to products and stock items)
    */
  public void registerSale(Sale sale)
  throws InventoryUnreachableException {
	logger.trace("Registering sale.");

    if (sale == null) {
      logger.trace("Sale is null, returning.");
      return;
    }

    int attemptsFailedWithInvalidContext = 0;
    int attemptsFailedWithTXAborted = 0;
    boolean done = false;

    // assert the component is in the correct state (throw exception otherwise)
    isActiveOrThrow();

    try {
      while (!done) {
    	logger.trace("About to start transaction.");
        try {
    
          // update stock items
          for(SaleItem si: sale.getSaleItems()) {
            // update stock item and save it
            updateStockItem(si);
            // rebind product
            rebindProduct(si);
          }

          // save the sale
          database.store(sale);

        } catch (InvalidContextException ice) {
          ++attemptsFailedWithInvalidContext;
          if (attemptsFailedWithInvalidContext >= INVALID_CONTEXT_RETRIES) {
            logger.warn("Register sale request denied because of permantly invalid request.");
            throw new InventoryUnreachableException("Giving up beacuse of permanently invalid context.", ice);
          }
        } catch (TransactionAbortedException tae) {
          ++attemptsFailedWithTXAborted;
          logger.trace("Transaction has been borted.");
          if (attemptsFailedWithTXAborted >= TRANSACTION_ABORTED_RETRIES) {
            logger.warn("Register sale request denied because transactions permanently abort.");
            throw new InventoryUnreachableException("Giving up because of permanently aborted transactions.");
          }
        } finally {
          done = database.commit();
          logger.trace("Transaction " + (done ? "successfully commited." : "aborted on commit."));
        }
      }
    } catch (DatabaseUnreachableException due) {
      logger.warn("Database unreachable for procesing register sale request.");
      throw new InventoryUnreachableException("Database unreachable for processing the request.", due);
    }
  }

  //SOFALifecycle methods

  /** Initializes this component.
    */
  public void start() {
    isActive = true;
  }

  /** Prepare for stopping this component. */
  public void stopping() {
    isActive = false;
  }

  /** Stop this component. */
  public void stop() {
  }

  // SOFAClient methods

  /** Sets requred interface endpoint.
    *
    * @param name name of the interface
    * @param object implementing instance
    */
  public void setRequired(String name, Object iface) {
    if (name.equals("data")) {
      if (iface instanceof ITransact) {
        database = (ITransact) iface;
      } else {
        throw new RuntimeException("Endpoint of required interface 'data' does not implement interface ITransact.");
      }
    }
    // ignore other interfaces
  } 

}
