/* $Id: ReportServer.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.rep;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.SOFAClient;
import org.objectweb.dsrg.sofa.demos.sofashop.db.ISelect;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.InvalidReportParameterException;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.Report;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportDataProcessingException;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportDataProvider;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportPrint;

/**
 * @author Ondrej Cerny
 *
 */
public class ReportServer implements IReporting, SOFAClient {
	
	private static Logger logger = Logger.getLogger(ReportServer.class);
	private static final String REQUIRED_ISELECT_INTERFACE_NAME = "select";
	
	/** Data provider. */
	private ReportDataProvider dataProvider;
	/** Data provider platform. */
	private ISelect database = null;
	/** Map of active reports. */
	private Map<Integer, Report> activeReportsMap = new HashMap<Integer, Report>();
	/** Report factory. */
	private ReportSelector reports = new ReportSelector();
	/** Last used report id. */
	private int lastId = 0;
	
	private int getNextId() {
		int id = 0;
		synchronized (this) {
			if (activeReportsMap.isEmpty()) {
				lastId = 0;
			}
			while (activeReportsMap.containsKey(Integer.valueOf(++lastId))) {
				if (lastId == Integer.MAX_VALUE) {
					lastId = 0;
					break;
				}
			}
			id = lastId;
		}
		return id;
	}
	
	private boolean pushActiveReport(int id, Report report)  {
		Integer i = new Integer(id);
		boolean result = false;
		synchronized (this) {
			if (activeReportsMap.containsKey(i)) {
				logger.error("Active reports map already contains report with id " + i.toString());
			} else {
				activeReportsMap.put(i, report);
				result = true;
			}
		}
		return result;
	}
	
	private Report popActiveReport(int id)  {
		Integer i = new Integer(id);
		Report result = null;
		synchronized (this) {
			if (activeReportsMap.containsKey(i)) {
				result = activeReportsMap.get(i);
				activeReportsMap.remove(i);
			} else {
				logger.error("Active reports map does not contain id " + i.toString());
			}
		}
		return result;
	}
	
	private ReportDataProvider getReportDataProvider() {
		ReportDataProvider result = null;
		synchronized (this) {
			if (dataProvider == null) {
				if (database == null) {
					throw new NullPointerException("Null ISelect provider.");
				}
				dataProvider = new ReportDataProviderImpl(database);
			}
			result = dataProvider;
		}	
		return result;
	}
	
	
	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.ReportingService#closeReport(int)
	  */
	@Override
	public void closeReport(int id) {
		popActiveReport(id);
	}
	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.ReportingService#getReportNames()
	  */
	@Override
	public Set<String> getReportNames() {
		return reports.getReportNames();
	}
	
	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.ReportingService#getReportParameters(int)
	  */
	@Override
	public List<ReportParameter> getReportParameters(int id)
	throws ReportDataProcessingException {
		List<ReportParameter> result = null;
		Report r = null;
		try {
			r = popActiveReport(id);
			if (r != null) {
				r.setDataProvider(getReportDataProvider());
				result = r.getParameters();
			}
		} catch (ReportDataProcessingException rdpe) {
			logger.info("Report data processing exception occured: " +rdpe.getMessage());
			throw rdpe;
		} catch (Throwable t) {
			logger.error("Unexpected error occured: ", t);
		} finally {
			if (r != null) {
				pushActiveReport(id, r);
			}
		}
		return result;
	}

	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.ReportingService#getReportParameters(int, java.util.List)
	  */
	@Override
	public List<ReportParameter> getReportParameters(int id,
			List<ReportParameter> currentValues)
	throws ReportDataProcessingException {
		List<ReportParameter> result = null;
		Report r = null;
		try {
			r = popActiveReport(id);
			if (r != null) {
				r.setDataProvider(getReportDataProvider());
				result = r.getParameters(currentValues);
			}
		} catch (ReportDataProcessingException rdpe) {
			logger.info("Report data processing exception occured: " +rdpe.getMessage());
			throw rdpe;
		} catch (Throwable t) {
			logger.error("Unexpected error occured: ", t);
		} finally {
			if (r != null) {
				pushActiveReport(id, r);
			}
		}
		return result;
	}

	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.ReportingService#getReportPrint(int, java.util.List)
	  */
	@Override
	public ReportPrint getReportPrint(int id, List<ReportParameter> values)
	throws ReportDataProcessingException, InvalidReportParameterException {
		ReportPrint result = null;
		Report r = null;
		try {
			r = popActiveReport(id);
			if (r != null) {
				r.setDataProvider(getReportDataProvider());
				result = r.getPrint(values);
			}
		} catch (ReportDataProcessingException rdpe) {
			logger.info("Report data processing exception occured: " +rdpe.getMessage());
			throw rdpe;
		} catch (InvalidReportParameterException irpe) {
			logger.info("Report parameter invalid: " + irpe.getMessage());
			throw irpe;
		} catch (Throwable t) {
			logger.error("Unexpected error occured: ", t);
		} finally {
			if (r != null) {
				pushActiveReport(id , r);
			}
		}
		return result;
	}

	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.ReportingService#openReport(java.lang.String)
	  */
	@Override
	public int openReport(String reportName) {
		int result = 0;
		Report r = reports.getReportByName(reportName);
		if (r != null) {
			for (result = getNextId(); result > 0 && !pushActiveReport(result, r); result = getNextId()) {}
		}
		return result;
	}
	/**
	  * 
	  * @see org.objectweb.dsrg.sofa.SOFAClient#setRequired(java.lang.String, java.lang.Object)
	  */
	@Override
	public void setRequired(String arg0, Object arg1) {
		if (arg0 == null || arg1 == null) {
			logger.error("setRequired(null, null)");
		} else if (REQUIRED_ISELECT_INTERFACE_NAME.equals(arg0)) {
			if (arg1 instanceof ISelect) {
				database = (ISelect) arg1;
			} else {
				logger.error("setRequired: '" + REQUIRED_ISELECT_INTERFACE_NAME +
						"' provider not instance of ISelect.");
			}
		} else {
			logger.error("setRequired: unkown interface name '" + arg0 +"'");			
		}
	}
}
