/* $Id: ReportParameter.java 770 2009-10-19 10:30:55Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.rep.common;

import java.io.Serializable;
import java.util.List;

/** Provides sudpplying parameter values to report.
  *
  * <p>
  * There can be identified three parameter states,
  * which should be considered when supplying report parameters.
  * 
  * <ul>
  * <li>parameter value is valid</li>
  * <li>parameter value has not been set yet</li>
  * <li>parameter value has not been set yet
  *     and it has no suitable available value
  *     (list of available values is empty) -
  *     this often means that there are outstanding
  *     dependencies (some other parameters must be set
  *     before)</li>
  * </li>
  * </p>
  * 
  * @author Ondrej Cerny
  *
  */
public interface ReportParameter extends Serializable {
	
	/** Returns parameter name.
	  * 
	  * <p>
	  * The parameter name uniquely identifies report parameter.
	  * </p>
	  * 
	  * @return parameter name
	  */
	public String getName();
	
	/** Returns parameter prompt.
	  * 
	  * <p>
	  * The parameter prompt represent human readable parameter identification.
	  * The prompt should be used to prompt user for supplying the value.
	  * </p>
	  * 
	  * @return parameter prompt
	  */	
	public String getPrompt();
	
	/** Returns parameter default value (if any).
	  * 
	  * @return default value or null
	  * @see #hasDefaultValue()
	  */
	public ReportParameterLabeledValue getDefaultValue();
	
	/** Returns current parameter value.
	  * 
	  * @return labeled value
	  */
	public ReportParameterLabeledValue getValue();
	
	/** Changes current parameter value.
	  * 
	  * @param value new labeled value of parameter
	  */
	public void setValue(ReportParameterLabeledValue value);
	
	/** Is null value allowed?.
	  * 
	  * @return is null value allowed?
	  */
	public boolean isNullAllowed();
	
	/** Is current parameter value valid?.
	  * 
	  * @return is current parameter value valid?
	  */
	public boolean isValueValid();
	
	/** Has the parameter a default value?
	  * 
	  * @return has the parameter a default value? 
	  */
	public boolean hasDefaultValue();
	
	
	/** Has the parameter any available values?
	  * 
	  * @return Has the parameter any available values?
	  */
	public boolean hasAvailableValues();

	/** Available values.
	  * 
	  * @return list of available values
	  */	
	public List<ReportParameterLabeledValue> getAvailableValues();
	
	/** Returns parameter type.
	  * 
	  * @return parameter type
	  */
	public Class<? extends Serializable> getType();
}
