/* $Id: ReportParameterImpl.java 770 2009-10-19 10:30:55Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.rep.common;

import java.io.Serializable;
import java.util.List;

/** Implementation of {@link ReportParameter}.
  * 
  * @author Ondrej Cerny
  *
  */
public class ReportParameterImpl extends AbstractReportParameter {
	
	protected ReportParameterLabeledValue currentValue = null;
	protected List<ReportParameterLabeledValue> availableValues = null;
	
	/** Creates instance of this class.
	  * 
	  * @param name parameter name
	  * @param prompt parameter prompt
	  * @param type parameter type
	  * @param isNullAllowed is null value allowed?
	  * @param defaultValue default value or null
	  */
	public ReportParameterImpl(String name, String prompt,
			Class<? extends Serializable> type,			
			boolean isNullAllowed,
			ReportParameterLabeledValue defaultValue) {
		super(name, prompt, type, isNullAllowed, defaultValue);
		if (hasDefaultValue()) {
			currentValue = getDefaultValue();
		}
	}
	

	/** Returns available values.
	  *
	  * @return {@link #availableValues} field
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter#getAvailableValues()
	  */
	@Override
	public List<ReportParameterLabeledValue> getAvailableValues() {
		return availableValues;
	}


	/** Returns current value.
	  * 
	  * @return {@link #currentValue}
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter#getValue()
	  */
	@Override
	public ReportParameterLabeledValue getValue() {
		return currentValue;
	}

	/** Has this parameter any available values?.
	  *
	  * @return true if and only if the {@link #availableValues}
	  *         field is not <code>null</code>
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter#hasAvailableValues()
	  */
	@Override
	public boolean hasAvailableValues() {
		return availableValues != null;
	}

	/** Examines whether the value is valid and returns the result.
	  * 
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter#isValueValid()
	  */
	@Override
	public boolean isValueValid() {
		if (currentValue == null) {
			return false;
		}
		if (currentValue.getValue() == null
				&& !isNullAllowed()) {
			return false;
		}
		if (hasAvailableValues() && availableValues == null) {
			// impossible with this implementation, but for subclasses
			return false;
		} 
		if (hasAvailableValues() && availableValues != null) {
			Serializable currentValueValue = currentValue.getValue();
			if (currentValueValue == null) {
				// search for null valued entry 
				for (ReportParameterLabeledValue rplv: availableValues) {
					if (rplv == null || rplv.getValue() == null) {
							return true;
					}
				}
			} else {
				// search for equal value
				for (ReportParameterLabeledValue rplv: availableValues) {
					if (currentValueValue.equals(rplv.getValue())) {
						return true;
					}
				}
			}
		}
		// no constraints violated
		return true;
	}

	/** Sets current parameter value.
	  *
	  * <p>
	  * If the parameter <i>value</i> is null, the {@link #defaultValue} will be
	  * used. If {@link #hasDefaultValue()} is false in such case,
	  * {@link ReportParameterLabeledValue} with <code>null</code>
	  * value and blank label will be used.
	  * </p>
	  *   
	  * @param value labeled value
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter#setValue(org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameterLabeledValue)
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameterLabeledValue
	  */
	@Override
	public void setValue(ReportParameterLabeledValue value) {
		if (value == null) {
			if (hasDefaultValue()) {
				currentValue = getDefaultValue();
			} else {
				currentValue = new ReportParameterLabeledValue(null, "");
			}
		} else {
			currentValue = value;
		}
	}
	
	/** Sets availabe values.
	  * 
	  * @param availableValues list of available values or null
	  *                        (null unsets available values).
	  */
	public void setAvailableValues(List<ReportParameterLabeledValue> availableValues) {
		this.availableValues = availableValues;
	}
}
