/* $Id: ParameterInput.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.rep.common.util;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.Serializable;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ComboBoxModel;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.TitledBorder;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameterLabeledValue;

/**
 * @author Ondrej Cerny
 *
 */
public class ParameterInput extends JPanel implements FocusListener {
	
	private static Logger logger = Logger.getLogger(ParameterInput.class);
	
	private ReportParameter p;
	private ReportParameterIntputContainerPane parent;
	private Component input = null;
	private boolean isDisplayed = false;
	
	public ParameterInput(ReportParameterIntputContainerPane parent) {
		this.parent = parent;
		this.setBorder(BorderFactory.createTitledBorder(""));
		this.setLayout(new GridBagLayout());
	}
	
	/** Updates this to reflect given parameter.
	  * 
	  * @param p parameter
	  */
	public void update(ReportParameter p) {
		if (!inputMatchesType(p)) {
			removeInput();
		}
		if (input == null) {
			createInput(p);
		}
		updateInput(p);
		updateTitle(p);
		this.validateTree();
		this.p = p;
	}
	
	private void updateTitle(ReportParameter p) {
		TitledBorder tb = (TitledBorder)this.getBorder();
		tb.setTitle(p != null && p.getPrompt() != null ? p.getPrompt() : "");
	}
	private boolean inputMatchesType(ReportParameter p) {
		if (input == null || p == null) {
			return false;
		}
		return (p.hasAvailableValues() && input instanceof JComboBox)
			||
				(!p.hasAvailableValues() && input instanceof JTextField);
	}	
	private void removeInput() {
		if (input != null) {
			this.remove(input);
		}
	}	
	private void createInput(ReportParameter p) {
		GridBagConstraints c = new GridBagConstraints();
		if (p.hasAvailableValues()) {
			input = new JComboBox();
		} else {
			input = new JTextField();
			c.ipadx = 100;			
		}
		input.addFocusListener(this);
		c.anchor = GridBagConstraints.WEST;
		c.fill = GridBagConstraints.HORIZONTAL;		
		this.add(input, c);
		
	}
	/** Updates input widget state.
	  * 
	  * <p>
	  * This action is initiated from outside and
	  * </p>
	  * @param p
	  */
	private void updateInput(ReportParameter p) {
		if (input != null && p != null) {
			ReportParameterLabeledValue val = p.getValue();
			
			if (input instanceof JComboBox) {
				logger.trace("Updating COMBO-BOX input with parameter " + p.getName() != null ? p.getName() : "<null-name>");
				JComboBox c = (JComboBox)input;
				// update available values
				ComboBoxModel m = c.getModel();
				DefaultComboBoxModel model = null;
				if (m instanceof DefaultComboBoxModel) {
					model = (DefaultComboBoxModel)m;
				}
				int index = 0;
				List<ReportParameterLabeledValue> vals = p.getAvailableValues();
				if (vals == null) {
					vals = new LinkedList<ReportParameterLabeledValue>();
				}
				if (model != null) {
					// compare the model against current values
					for (ReportParameterLabeledValue v : vals) {
						if (!v.equals(model.getElementAt(index))) {
							model = null;
							break;
						}
						++index;
					}
				}
				if (model == null) {
					logger.trace("Updating COMBO-BOX input with parameter " + p.getName() != null ? p.getName() : "<null-name>" + " - new available values");					
					model = new DefaultComboBoxModel(vals.toArray());
				}
				c.setModel(model);
								
				// udpate selection
				c.setSelectedItem(p.getValue());
				
			} else if (input instanceof JTextField) {
				logger.trace("Updating TEXT-FIELD input with parameter " + p.getName() != null ? p.getName() : "<null-name>");
				JTextField field = (JTextField)input;
				if (val == null) {
					field.setText("");
				} else {
					field.setText(val.toString());
				}
			} else {
				logger.error("Internal: incompatible input component.");
			}
		} else {
			logger.warn("Attempted to update null input or with null parameter. Ignoring.");
		}
	}

	/** Handles focus gained event.
	  *
	  * <p>
	  * Does nothing.
	  * </p>
	  * 
	  * @param e focus event
	  */
	@Override
	public void focusGained(FocusEvent e) {
	}

	/** Handles focus lost event of the input widget.
	  *
	  * <p>
	  * Focus-lost event is a good time to update parameter value.
	  * Updates {@link #p parameter} internally and then calls 
	  * {@link ReportParameterIntputContainerPane#parameterSelectionChanged(ReportParameter)}
	  * with it on {@link #parent}.
	  * </p>
	  * 
	  * @param e focus event
	  */
	@Override
	public void focusLost(FocusEvent e) {
		logger.trace("Focus lost at parameter input for name " + p != null && p.getName() != null ? p.getName() : "<null-name>");
		// update selected value...
		if (p != null && input != null) {
			if (input instanceof JComboBox) {
				logger.trace("About to update current value of parameter (COMBO-BOX).");
				Object o = ((JComboBox)input).getSelectedItem();
				ReportParameterLabeledValue currVal = null;
				if (o instanceof ReportParameterLabeledValue) {
					currVal = (ReportParameterLabeledValue)o;
				}
				if (currVal != null) {
					p.setValue(currVal);
					parent.parameterSelectionChanged(p);					
				} else {
					// reselect back
					((JComboBox)input).setSelectedItem(p.getValue());
				}
			} else if (input instanceof JTextField) {
				logger.trace("About to update current value of parameter (TEXT-FIELD).");
				JTextField tf = (JTextField)input;
				Class<? extends Serializable> c = p.getType();
				String label = tf.getText();
				Serializable value = ParameterValueConversionHelper.valueOf(c, label);
				ReportParameterLabeledValue newVal = new ReportParameterLabeledValue(value, label);
				p.setValue(newVal);
				parent.parameterSelectionChanged(p);
			} else {
				logger.error("Internal: Unrecognized parameter input component type.");
			}
		}
	}
	/** Marks this as (not) displayed.
	  * 
	  * <p>
	  * Just sets {@link #isDisplayed} attribute.
	  * </p>
	  * 
	  * @param displayed value
	  */
	public void setIsDisplayed(boolean displayed) {
		isDisplayed = displayed;
	}
	/** Returns indication whether this has been marked as displayed.
	  *
	  * <p>
	  * Just returns value of {@link #isDisplayed} attribute.
	  * </p>
	  * 
	  * @return value of {@link #isDisplayed}
	  */
	public boolean isDisplayed() {
		return isDisplayed;
	}
	
	/** Sets {@link #input} enabled.
	  * 
	  * @param value enabled
	  */
	public void setControlEnabled(boolean value) {
		if (input != null) {
			input.setEnabled(value);
		}
	}
}
