/* $Id: OverviewOfOrdersReport.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.rep.samplereports;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.neodatis.odb.core.query.IQuery;
import org.neodatis.odb.core.query.criteria.ComposedExpression;
import org.neodatis.odb.core.query.criteria.Where;
import org.neodatis.odb.impl.core.query.criteria.CriteriaQuery;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.OrderItem;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Supplier;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.DefaultTextReportPrint;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.InvalidReportParameterException;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.Report;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportDataProcessingException;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportDataProvider;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameterImpl;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameterLabeledValue;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportPrint;

/**
 * @author Ondrej Cerny
 *
 */
public class OverviewOfOrdersReport implements Report {
	
	private static Logger logger = Logger.getLogger(OverviewOfOrdersReport.class);
	private static final String ORDER_DATE_FROM_PARAM_NAME = "from";
	private static final String ORDER_DATE_TO_PARAM_NAME = "to";
	private static final String ORDER_STATUS_PARAM_NAME = "status";
	
	private static final int ORDER_STATUS_ALL = 0;
	private static final int ORDER_STATUS_OPEN = 1;
	private static final int ORDER_STATUS_DELIVERED = 2;
	
	
	ReportDataProvider temporaryDataProvider = null;
	List<ReportParameter> parameters;
	
	public OverviewOfOrdersReport() {
		parameters = new LinkedList<ReportParameter>();		
		parameters.add(new ReportParameterImpl(
				ORDER_DATE_FROM_PARAM_NAME,
				"From", java.util.Date.class, false, null));
		parameters.add(new ReportParameterImpl(
				ORDER_DATE_TO_PARAM_NAME,
				"To", java.util.Date.class, false, null));
		ReportParameterImpl status = new ReportParameterImpl(ORDER_STATUS_PARAM_NAME,
				"State", Integer.class, false, null);
		List<ReportParameterLabeledValue> availableStatuses = new LinkedList<ReportParameterLabeledValue>();
		availableStatuses.add(new ReportParameterLabeledValue(Integer.valueOf(ORDER_STATUS_ALL), "All"));
		availableStatuses.add(new ReportParameterLabeledValue(Integer.valueOf(ORDER_STATUS_OPEN), "Not delivered yet"));
		availableStatuses.add(new ReportParameterLabeledValue(Integer.valueOf(ORDER_STATUS_DELIVERED), "Delivered"));
		status.setAvailableValues(availableStatuses);
		parameters.add(status);
	}
	
	@Override
	public List<ReportParameter> getParameters() {
		return parameters;
	}

	@Override
	public List<ReportParameter> getParameters(
			List<ReportParameter> currentValues) {
		return currentValues;
	}

	@Override
	public ReportPrint getPrint(List<ReportParameter> values)
			throws InvalidReportParameterException, ReportDataProcessingException {
		java.util.Date from = null;
		java.util.Date to = null;
		Integer status = null;
		try {
			for (ReportParameter p: values) {
				if (ORDER_DATE_FROM_PARAM_NAME.equals(p.getName())) {
					from = (java.util.Date)p.getValue().getValue();
				} else if (ORDER_DATE_TO_PARAM_NAME.equals(p.getName())) {
					to = (java.util.Date)p.getValue().getValue();
				} else if (ORDER_STATUS_PARAM_NAME.equals(p.getName())) {
					status = (Integer)p.getValue().getValue();				
				} else {
					logger.trace("Unknown parameter: " + (p.getName() != null ? p.getName() : "<null-name>"));
				}
			}
		} catch (Throwable t) {
			logger.trace("Error occured when parsing parameters: " + t.getClass().getName() + ": " + t.getMessage());
		} finally {
			if (from == null) {
				throw new InvalidReportParameterException(
						"Wrong value supplied for parameter " +
						ORDER_DATE_FROM_PARAM_NAME + ".");
			}
			if (to == null) {
				throw new InvalidReportParameterException(
						"Wrong value supplied for parameter " +
						ORDER_DATE_TO_PARAM_NAME + ".");				
			}
			if (status == null) {
				throw new InvalidReportParameterException(
						"Wrong value supplied for parameter " +
						ORDER_STATUS_PARAM_NAME + ".");								
			}
			if (!from.before(to)) {
				throw new InvalidReportParameterException(
						"Value of parameter " + ORDER_DATE_FROM_PARAM_NAME +
						" must be less than " + ORDER_DATE_TO_PARAM_NAME + ".");
			}
		}
		
		// we have parsed parameters - lets prepare query
		ComposedExpression andCrit = Where.and().
			add(Where.ge("createDate", from)).
			add(Where.le("createDate", to));
		int statusVal = status.intValue();		
		switch (statusVal) {
		case ORDER_STATUS_ALL:
			break;
		case ORDER_STATUS_DELIVERED:
			andCrit.add(Where.isNotNull("closeDate"));
			break;
		case ORDER_STATUS_OPEN:
			andCrit.add(Where.isNull("closeDate"));
			break;
		default:
			throw new InvalidReportParameterException("Value of parameter " +
					ORDER_STATUS_PARAM_NAME + " is invalid.");
		}
		IQuery q = new CriteriaQuery(Order.class, andCrit);
		
		// query created - lets select data 
		Collection<Order> orders = temporaryDataProvider.select(q);
		
		// lets build html code
		StringBuilder sb = new StringBuilder();
		
		sb.append("<body>");
		sb.append("<h1>Overview of orders</h1>");
		
		if (orders != null && orders.size() > 0) {
			sb.append("<table>");
			
			sb.append("<tr>");
			sb.append("<td><b>Order code</b></td>");
			sb.append("<td><b>Supplier name</b></td>");
			sb.append("<td><b>Supplier reg. no.</b></td>");
			sb.append("<td><b>Created</b></td>");
			sb.append("<td><b>Closed</b></td>");
			sb.append("<td><b>Total quantity</b></td>");
			sb.append("</tr>");
			
			for (Order o: orders) {
				if (o == null) {
					continue;
				}
				sb.append("<tr>");
				sb.append("<td>").append(o.getCode() != null ? o.getCode() : "---").append("</td>");
				Supplier s = o.getSupplier();
				sb.append("<td>").append(s != null && s.getName() != null ? s.getName() : "---").append("</td>");
				sb.append("<td>").append(s != null && s.getRegNo() != null ? s.getRegNo() : "---").append("</td>");
				sb.append("<td>").append(o.getCreateDate() != null ? o.getCreateDate().toString() : "---").append("</td>");
				sb.append("<td>").append(o.getCloseDate() != null ? o.getCloseDate().toString() : "---").append("</td>");
				double amount = 0;
				if (o.getOrderItems() != null) {
					for (OrderItem oi: o.getOrderItems()) {
						if (oi != null) {
							amount += oi.getAmount();
						}
					}
				}
				sb.append("<td>").append(amount).append("</td>");
				sb.append("</tr>");				
			}
			
			
			sb.append("</table>");						
		} else {
			sb.append("<p>No such order.</p>");
		}
		
		sb.append("</body>");		
		return new DefaultTextReportPrint("text/html", sb.toString());
	}

	@Override
	public void setDataProvider(ReportDataProvider p) {
		temporaryDataProvider = p;
	}

}
