/* $Id: InventoryDataHelper.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.store.client;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Set;
import java.util.concurrent.locks.Lock;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.StockItem;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.order.IOrder;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.order.OrderInconsistentException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.prod.IListProducts;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.Report;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.service.AbstractReportingServiceClient;
import org.objectweb.dsrg.sofa.demos.sofashop.store.client.debug.TestingProductsDataSource;

/**
 * @author Ondrej Cerny
 *
 */
public abstract class InventoryDataHelper {
	private static Logger logger = Logger.getLogger(InventoryDataHelper.class);
	
	private static InventoryResourceManager rm = new InventoryResourceManager();
	private static volatile IListProducts productsList = null;
	private static volatile IOrder orderRegistry = null;
	private static volatile AbstractReportingServiceClient RSC = null;
	
	public static synchronized void setProductsList(IListProducts newProductsList) {
		logger.trace("->");								
		if (productsList != null) {
			logger.fatal("Initializing InventoryDataHelper: " +
					"setting products list twice. Ignoring");
			return;
		}
		productsList = newProductsList;
		rm.releaseResource(0);
		logger.trace("->");										
	}	
	private static IListProducts getProductsList() {
		logger.trace("->");
		rm.acquireResource(0);
		logger.trace("<-");		
		return productsList;
	}
	private static void releaseProductsList() {
		logger.trace("->");		
		rm.releaseResource(0);
		logger.trace("<-");				
	}
	public static synchronized void setOrderRegistry(IOrder newOrderRegistry) {
		logger.trace("<-");												
		if (orderRegistry != null) {
			logger.fatal("Initializing InventoryDataHelper: " +
				"setting order registry twice. Ignoring");								
			return;
		}
		orderRegistry = newOrderRegistry;
		rm.releaseResource(1);
		logger.trace("<-");												
	}	
	private static IOrder getOrderRegistry() {
		logger.trace("->");				
		rm.acquireResource(1);
		logger.trace("<-");						
		return orderRegistry;
	}
	private static void releaseOrderRegistry() {
		logger.trace("->");						
		rm.releaseResource(1);
		logger.trace("<-");								
	}
	
	public static synchronized void setRSC(AbstractReportingServiceClient newRSC) {
		logger.trace("<-");												
		if (orderRegistry != null) {
			logger.fatal("Initializing InventoryDataHelper: " +
				"setting order registry twice. Ignoring");								
			return;
		}
		RSC = newRSC;
		rm.releaseResource(2);
		logger.trace("<-");												
	}	
	private static AbstractReportingServiceClient getRSC() {
		logger.trace("->");				
		rm.acquireResource(2);
		logger.trace("<-");						
		return RSC;
	}
	private static void releaseRSC() {
		logger.trace("->");						
		rm.releaseResource(2);
		logger.trace("<-");								
	}
	
	
	
	/** Returns collection of {@link StockItems} ordered by {@link Product product} id.
	  *
	  * <p>
	  * For single product, just one stock item will be present in the result.
	  * For products with different number of stock items than one,
	  * an extra stock item will be created (it will represent the total
	  * number of items on the stock).   
	  * </p>
	  * 
	  * <p>
	  * Note that the stock item may be created especially for this purpose
	  * thus it must not be stored on database.
	  * </p>
	  * 
	  * 
	  * @param from index of the first returned element from the global order
	  * @param amount maximal number of returned elements (-1 for all)
	  * @return collection of {@link StockItem stock items}
	  * @see Product#getId()
	  * @see Product#setId(long)
	  */
	public static Collection<Product> getProducts(
			String namePattern,
			int from, int amount)
			throws InventoryUnreachableException {
		Collection<Product> result = null;
		IListProducts pl = getProductsList();
		try {
			result = pl.listProducts(namePattern, from,
					(amount > 0 ? from + amount : -1));
		} finally {
			releaseProductsList();
		}
		return result;
	}
	/** Returns product having given product id.
	  * 
	  * @param productId product id
	  * @return {@link Product}
	  * @see Product#getId()
	  * @see Product#setId(long)
	  */
	public static Product getProductForId(long productId)
	throws InventoryUnreachableException {
		Product p = null;
		IListProducts pl = getProductsList();		
		try {
			p = pl.getProductById(productId);
		} finally {
			releaseProductsList();
		}
		return p;
	}
	
	/** Saves order to inventory.
	  * 
	  * @param order
	  * @return report code or null if the order has not been commited correctly
	  * @throws OrderInconsistentException
	  * @throws InventoryUnreachableException
	  * @see IOrder#registerOrder(Order) 
	  */
	public static String saveOrder(Order o) 
	throws OrderInconsistentException, InventoryUnreachableException {
		String orderCode = null;
		IOrder orderRegistry = getOrderRegistry();
		try {
			orderCode = orderRegistry.registerOrder(o);
		} finally {
			releaseOrderRegistry();
		}
		return orderCode;
	}	
	
	/** Closes previously created order.
	  * 
	  * <p>
	  * Closing order means that the goods has been received,
	  * so appropriate stock items will be created as well
	  * as the order will be marked as closed.
	  * </p>
	  *  
	  * @param orderCode
	  * @throws OrderInconsistentException
	  * @throws InventoryUnreachableException
	  * @see IOrder#receiveOrder(Order)
	  */
	public static void closeOrder(String orderCode)
	throws OrderInconsistentException, InventoryUnreachableException {
		if (orderCode == null || orderCode.equals("")) {
			throw new OrderInconsistentException(
					"Null or empty order code.");
		}
		IOrder orderRegistry = getOrderRegistry();
		try {
			orderRegistry.receiveOrder(orderCode);
		} finally {
			releaseOrderRegistry();
		}
	}
	
	public static Set<String> getReportNames() {
		Set<String> result = null;
		AbstractReportingServiceClient rsc = getRSC();
		try {
			logger.trace("Calling report server for all available report names.");
			result = rsc.getReportNames();
		} finally {
			releaseRSC();
		}
		return result;
	}
	
	public static Report getReportForName(String name) {
		if (name == null) {
			logger.trace("Will not bother report serve with null name.");
			return null;
		}
		Report result = null;
		AbstractReportingServiceClient rsc = getRSC();
		try {
			logger.trace("Calling report server for report named " + name + ".");			
			result = rsc.getReportForName(name);
		} finally {
			releaseRSC();
		}
		return result;
	}	
}
