/* $Id: OrderItemDSTTableModel.java 806 2009-11-15 10:08:58Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.store.client.gui.order;

import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Supplier;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.SupplyTerms;

/** Product's supply terms table model.
  * 
  * @author Ondrej Cerny
  *
  */
public class OrderItemDSTTableModel extends AbstractTableModel implements TableModel {
	
	/** Logger. */
	private static Logger logger = Logger.getLogger(OrderItemDSTTableModel.class);

	private static final String[] columnNames = {"Supplier name", "Reg. No.", "Valid From" , "Valid To", "Qty. From", "Qty. To", "Unit price"};
	
	private List<SupplyTerms> wholeList = new LinkedList<SupplyTerms>();
	private List<SupplyTerms> filteredList = new LinkedList<SupplyTerms>();
	private boolean isFiltered = true;	
	private List<SupplyTerms> currentList = null;
	private double amount = 0;
	private boolean isSortedBySupplier = false;
	/** For sorting supply terms by price.
	  */
	private Comparator<SupplyTerms> supplyTermsComparator =
		new Comparator<SupplyTerms>() {
		/** Compares two supply terms by price.
		  * 
		  * @param o1 
		  * @param o2
		  * @return negative integer, zero or positive integer as the o1's price is
		  *         less than, equal or greater than o2's price 
		  */
		@Override
		public int compare(SupplyTerms o1, SupplyTerms o2) {
			// push nulls back
			if (o1 == null) {
				return -1;
			} else if (o2 == null) {
				return +1;
			}
			// compare
			if (o1.getPrice() < o2.getPrice()) { 
				return -1;
			} else if (o1.getPrice() > o2.getPrice()) {
				return +1;
			}
			return 0;
		}		
	};
	/** For sorting supply terms by supplier name.
	  */
	private Comparator<SupplyTerms> supplyTermsComparatorBySupplierName = 
		new Comparator<SupplyTerms>() {
		/** Compares two supply terms by supplier name.
		  * 
		  * @param o1 
		  * @param o2
		  * @return negative integer, zero or positive integer as the o1's price is
		  *         less than, equal or greater than o2's price 
		  */
		@Override
		public int compare(SupplyTerms o1, SupplyTerms o2) {
			// push null objects back
			if (o1 == null) {
				return -1;
			} else if (o2 == null) {
				return +1;
			}
			
			String name1 = o1.getSupplier() != null
				? o1.getSupplier().getName() : null;
			String name2 = o2.getSupplier() != null
				? o2.getSupplier().getName() : null;
			
		    if (name1 == null) {
				return -1;
			} else if (name2 == null) {
				return +1;
			}
		
			// compare
			return name1.compareTo(name2); 
		}
	};

	/** Decides if given supply terms are applicable to given amount
	  * 
	  * @param supplyTerms supply terms
	  * @param amount amount
	  * @return
	  */
	public static boolean isSupplyTermsApplicableToAmount(
			SupplyTerms supplyTerms, double amount) {
		return (supplyTerms != null
				&& supplyTerms.getLowerLimitIncl() <= amount
				&& (supplyTerms.getUpperLimitExcl() > amount 
						||
						supplyTerms.getUpperLimitExcl() == -1));
	}
	/** Decides whether given supply terms are applicable to current date-time.
	  * 
	  * @param supplyTerms supply terms
	  * @return
	  */
	public static boolean isSupplyTermsApplicableRightNow(
			SupplyTerms supplyTerms) {
		Date now = new Date();
		Date from = supplyTerms != null ? supplyTerms.getValidFrom() : null;
		Date to = supplyTerms != null ? supplyTerms.getValidTo() : null;
		return (supplyTerms != null
				&& (from == null || from.before(now))
				&& (to == null || to.after(now)));
	}	
	/** Recomputes the filtered list.
	  *
	  * <p>
	  * <b>Must</b> be called whenever {@link OrderItemDSTTableModel#wholeList} or
	  * {@link #isFiltered} are changed.
	  * </p>
	  * 
	  * <p>
	  * Moreover, whenever this method is called, the {@link #wholeList} <b>must</b> be 
	  * already sorted to the final form.
	  * </p>
	  */
	protected void updateFilteredList() {
		filteredList.clear();
		if (isFiltered) {
			logger.trace("Rebuilding the filtered list.");
			for(SupplyTerms st : wholeList) {
				if (isSupplyTermsApplicableToAmount(st, amount)
						&& isSupplyTermsApplicableRightNow(st)) {
					filteredList.add(st);
				}					
			}
		}
	}
	protected void updateDataSource()  {
		currentList = isFiltered ? filteredList : wholeList;
	}
	protected void sortDataSource() {
		Comparator<SupplyTerms> cmp = isSortedBySupplier
				? this.supplyTermsComparatorBySupplierName 
				: this.supplyTermsComparator;
		
		logger.trace("Sorting the whole list.");
		Collections.sort(wholeList, cmp);
		if (isFiltered) {
			// sort the filtered list too (it is supposed to be at
			// most as hard as creating the filtered list from scratch
			logger.trace("Sorting the filtered list too.");				
			Collections.sort(filteredList, cmp);
		}
	}
	protected void causeRedrawTable() {
		fireTableDataChanged();
	}
	/** Copies supply terms list from {@link Product product}.
	  * 
	  * <p>
	  * Note that this class will very probably reorder and manipulate
	  * the list, that's why a copy is made here.
	  * </p>
	  * 
	  * @param p product
	  */
	public void dataValuesOf(Product p) {
		logger.trace("Getting values from product.");
		List<SupplyTerms> originalList = p != null
				? p.getSupplyTerms()
				: null;
		wholeList.clear();
		filteredList.clear();
		if (originalList != null) {
			for (SupplyTerms st: originalList) {
				if (st != null) {
					wholeList.add(st);
				}
			}
			updateFilteredList();
		}
		sortDataSource();
		updateDataSource();
		causeRedrawTable();
	}
	/** Sets whether to filter supply terms by applicability to current amount.
	  * 
	  * @param filter
	  */
	public void setIsFiltered(boolean filter) {
		if (filter != isFiltered) {
			logger.trace("Setting " + (filter ? "is " : "not ") + "filtered.");
			isFiltered = filter;
			updateFilteredList();
			updateDataSource();
			causeRedrawTable();
		}
	}
	/** Sets whether to sort by supply terms or by unit price.
	  * 
	  * @param bySupplier 
	  * 	if <code>true</code>, the table will be sorted by supplier name
	  *		or it will be sorted by unit price otherwise (which is default).
	  */
	public void setSortBySupplier(boolean bySupplier) {
		if (isSortedBySupplier != bySupplier) {
			logger.trace("Setting " + (bySupplier ? "" : "not ") +
					"sorted by supplier name.");
			isSortedBySupplier = bySupplier;
			
			sortDataSource();
			causeRedrawTable();
		} 
		// else do nothing
	}
	/** Sets the currently selected amount.
	  * 
	  * @param amount
	  */
	public void setAmount(double amount) {
		if (this.amount != amount) {
			logger.trace("Seting amount.");
			this.amount = amount;
			updateFilteredList();
			causeRedrawTable();
		}
	}
	
	/** Returns {@link SupplyTerms} object stated at given index. 
	  * 
	  * @param index index
	  * @return {@link SupplyTerms}
	  */
	public SupplyTerms getSupplyTerms(int index) {
		SupplyTerms result = null;
		if (index >= 0 && currentList != null && currentList.size() > index) {
			result = currentList.get(index);
		}
		return result;
	}
	

	@Override
	public int getColumnCount() {
		return columnNames.length;
	}

	@Override
	public int getRowCount() {
		if (currentList == null) {
			return 0;
		} else {
			return currentList.size();
		}
	}

	@Override
	public Object getValueAt(int rowIndex, int columnIndex) {
		SupplyTerms st = currentList != null ? currentList.get(rowIndex) : null;
		if (st == null) {
			return null;
		}
		Supplier s;
		Date d;
		double limit;
		switch (columnIndex) {
		case 0: s = st.getSupplier();
				return s != null ? s.getName() : null;
		case 1: s = st.getSupplier();
				return s != null ? String.format("%10s", s.getRegNo()) : null;
		case 2: d = st != null ? st.getValidFrom() : null;
				return d != null ? String.format("%tY/%tm/%td", d, d, d) : "----/--/--";
		case 3: d = st != null ? st.getValidTo() : null;
				return d != null ? String.format("%tY/%tm/%td", d, d, d) : "----/--/--";
		case 4: return Double.toString(st.getLowerLimitIncl());
		case 5: limit = st.getUpperLimitExcl();
				if (limit == -1) {
					return "*";
				} else {
					return String.format("%.2f", limit);
				}
		case 6: return String.format("%.2f", st.getPrice());
		default: return null;
		}
	}
	
	@Override
    public String getColumnName(int col) {
		return columnNames[col];
    }
	
	@Override
	public boolean isCellEditable(int row, int col) {
		return false;
    }

	
	
}
