/* $Id: OrderOverviewPane.java 762 2009-10-12 11:19:00Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.store.client.gui.order;

import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.concurrent.TimeUnit;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTree;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.OrderItem;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Product;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Supplier;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.SupplyTerms;
import org.objectweb.dsrg.sofa.demos.sofashop.store.client.gui.reports.ReportsPane;

/** Shows overview of current order and allows it to be submitted or canceled.
  * 
  * @author Ondrej Cerny
  */
public class OrderOverviewPane extends JPanel implements ActionListener {
	
	private static final String SUBMIT_OPTIONS_STRINGS[] = {"Yes, show reports, please", "Yes", "No"};	
	/** Logger. */
	private static Logger logger = Logger.getLogger(OrderOverviewPane.class);
	
	/** Views order reports. */
	private ReportsPane orderReportViewer;
	
	/** Root node. */
	private OrderTreeNode root = new OrderTreeNode();
	
	/** List of submitted orders.
	  *
	  * <p>
	  * Noted that the list will be available from order submit
	  * to creation of next order
	  * (which is done on {@link #addOrderItem(Product, SupplyTerms, double)}
	  * call).
	  * </p>
	  * 
	  */ 
	//private List<Order> submittedOrders = null;
	
	/** Returns root tree node.
	  * 
	  * @return {@link OrderTreeNode}
	  */
	private OrderTreeNode getRootNode() {
		return root;
	}

	/** Creates {@link OrderTreeNode tree node} with {@link Order order}
	  * with default field values except the {@link Supplier},
	  * which will be set immediately. 
	  * 
	  * @param s {@link SupplyTerms supply terms} 
	  * @return {@link OrderTreeNode}
	  */
	private OrderTreeNode createInitialOrder(Supplier s) {
		Order o = new Order();
		o.setSupplier(s);
		return new OrderTreeNode(o);
	}
	
	private void cleanOrderTree() {
		logger.trace("Cleaning order tree.");
		getRootNode().removeAllChildren();
		refreshOrderView();
	}
	
	private void refreshOrderView() {
		getTree().updateUI();
	}
	
	/** Adds new {@link OrderItem order item} to particular {@link Order order}.
	  * 
	  * @param o {@link OrderTreeNode order tree node}
	  *          carrying the {@link Order order}
	  * @param p {@link Product product}
	  * @param t {@link SupplyTerms supply terms}
	  * @param amount amount
	  * @return newly created order tree node
	  * @throws IllegalArgumentException invalid order
	  */
	private OrderTreeNode addOrderItemToOrder(OrderTreeNode o, Product p, SupplyTerms t, double amount)
	throws IllegalArgumentException {
		if (o == null || !o.isOrder()) {
			throw new IllegalArgumentException("Insufficient order or its handle.");
		}
				
		OrderItem oi = new OrderItem();
		oi.setProduct(p);
		oi.setTerms(t);
		oi.setAmount(amount);
		OrderTreeNode otn = new OrderTreeNode(oi);
		
		o.add(otn);
		logger.trace("Order item added.");
		return otn;
	}
	/** Constructor.
	  * 
	  * @param orderReportViewer views order reports
	  *                          (if null, the report display will not be allowed)
	  */
	public OrderOverviewPane(ReportsPane orderReportViewer) {
		super();
		this.orderReportViewer = orderReportViewer;
		// init swing components
		init();
	}
	
	/** Default constructor.
	  * 
	  */
	public OrderOverviewPane() {
		this(null);
	}
	
	/** Creates {@link OrderItem order item} for given
	  * {@link Product product}, {@link SupplyTerms supply terms}
	  * and amount and adds it to the view.
	  * 
	  * @param p {@link Product product}
	  * @param t {@link SupplyTerms supply terms}
	  * @param amount amount
	  * @throws IllegalArgumentException when the product, supply terms or amount
	  *                                  is insufficient
	  */
	public void addOrderItem(Product p, SupplyTerms t, double amount)
	throws IllegalArgumentException {
		boolean inserted = false;
		logger.trace("Adding order item.");
		//if (submittedOrders != null) {
			// new order is being created right now
			// clean old order data
		//	submittedOrders = null;
		//	cleanOrderTree();
		//}
		Supplier s = t != null ? t.getSupplier() : null;
		if (p == null || s == null || amount <= 0
				|| s.getRegNo() == null) {
			throw new IllegalArgumentException("Unable to add order item.");
		}
		OrderTreeNode root = getRootNode(); 
		int orders = root.getChildCount();
		for (int i = 0; i < orders; ++i) {
			TreeNode tn = root.getChildAt(i);
			OrderTreeNode otn = (OrderTreeNode)tn;
			if (!otn.isOrder()) {
				logger.warn("Unexpected type of tree node. Skipping.");
				break;
			}
			Order o = otn.getOrder();
			if (o.getSupplier().getRegNo().equals(s.getRegNo())) {
				logger.trace("Adding order item to existing Order.");				
				getTree().expandPath( new TreePath(
						addOrderItemToOrder(otn, p, t, amount)));

				inserted = true;
				break;
			}
		}
		if (!inserted) {
			logger.trace("Adding order item to new Order.");
			OrderTreeNode newOrder = createInitialOrder(s);
			OrderTreeNode tmp = addOrderItemToOrder(newOrder ,p, t, amount);
			root.add(newOrder);
			getTree().expandPath(new TreePath(tmp));
		
		}
		
		refreshOrderView();
		
	}
	
	// swing part:
	private JTree tree = null;
	private JButton submitButton = null;
	private JButton cancelButton = null;
	
	private void init() {
		setBorder(BorderFactory.createTitledBorder("Order"));
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		add(new JScrollPane(getTree()));
		
		JPanel ctrls = new JPanel();
		ctrls.setLayout(new FlowLayout(FlowLayout.RIGHT));
		ctrls.add(getCanceButton());		
		ctrls.add(getSubmitButton());
		add(ctrls);
	}
	private JTree getTree() {
		if (tree == null) {
			tree = new JTree(getRootNode());
			tree.setRootVisible(false);
		}
		return tree;
	}
	private JButton getSubmitButton() {
		if (submitButton == null) {
			submitButton = new JButton("Submit");
			submitButton.addActionListener(this);
		}
		return submitButton;
	}
	private JButton getCanceButton() {
		if (cancelButton == null) {
			cancelButton = new JButton("Clean");
			cancelButton.addActionListener(this);			
		}
		return cancelButton;
	}
	private void setControlsEnabled(boolean enabled) {
		getTree().setEnabled(enabled);
		getSubmitButton().setEnabled(enabled);
		getCanceButton().setEnabled(enabled);
	}
	private JPanel getThisPane() {
		return this;
	}
	
	/** Submits orders and saves them onto inventory.
	  *
	  * <p>
	  * Note that if some (or all) orders has not been 
	  * submitted successfully, they are left in the tree
	  * and  
	  * 
	  * @return true if and only if
	  */
	private void submitOrders(boolean tryShowRepors) {
		// disable controls
		setControlsEnabled(false);
		
		// create worker and execute
		OrderRegistryWorker orw = new OrderRegistryWorker(
					getTree(),
					tryShowRepors ? orderReportViewer : null) {
			@Override
			protected void done() {
				//get result
				int failed = -1;
				try {
					failed = get(10, TimeUnit.MILLISECONDS).intValue();
					// inform user about the result
					if (failed == 0) {
						JOptionPane.showMessageDialog(getThisPane(),
								"All orders submitted successfully.",
								"Submission", JOptionPane.INFORMATION_MESSAGE);						
					} else {
						JOptionPane.showMessageDialog(getThisPane(),
								"Some orders has not been registered successfully. " +
								"You can edit and/or submit them again.",
								"Submission", JOptionPane.WARNING_MESSAGE);
					}					
				} catch (Throwable t) {
					logger.error("An error occured in get() called from done()", t);
					JOptionPane.showMessageDialog(getThisPane(),
							"An internal error occured when fetching " +
							"order registration result, see log " +
							"for more details.", "Internal Error",
							JOptionPane.ERROR_MESSAGE);
				}
				
				// enable controls
				setControlsEnabled(true);
				
				// refresh tree view
				refreshOrderView();
			}
		};
		orw.execute();

	}

	// +---------------------+
	// +ActionListener method+
	// +---------------------+
	@Override
	public void actionPerformed(ActionEvent e) {
		Object source = e.getSource();
		if (source != null && source == getSubmitButton()) {
			// submit
			logger.trace("Submit pressed.");
 
			int result = JOptionPane.showOptionDialog(
					this, "Are you sure you want to submit all the orders?",
					"Submission", JOptionPane.YES_NO_CANCEL_OPTION,
					JOptionPane.QUESTION_MESSAGE, null,
					SUBMIT_OPTIONS_STRINGS, null); 
			logger.debug("Option: " + (result == JOptionPane.YES_OPTION ? "YES" : (result == JOptionPane.NO_OPTION ? "NO" : (result == JOptionPane.CANCEL_OPTION ? "CANCEL" : "unknown"))));
			if (result == JOptionPane.YES_OPTION || result == JOptionPane.NO_OPTION) {
				submitOrders(result == JOptionPane.YES_OPTION);
			}						
		} else if (source != null && source == getCanceButton()) {
			// cancel
			logger.trace("Clean pressed.");
			int result = JOptionPane.showConfirmDialog(
					this, "Are you sure you want clean all the orders?",
					"Clean Orders", JOptionPane.YES_NO_OPTION);
			if (result == JOptionPane.YES_OPTION) {
				cleanOrderTree();
			}			
		} else {
			logger.error("Unrecognized action event source component.");
		}
	}
}
