/* $Id: OrderRegistryWorker.java 762 2009-10-12 11:19:00Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.objectweb.dsrg.sofa.demos.sofashop.store.client.gui.order;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JTree;
import javax.swing.SwingWorker;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeModel;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.Order;
import org.objectweb.dsrg.sofa.demos.sofashop.common.datamodel.OrderItem;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.InventoryUnreachableException;
import org.objectweb.dsrg.sofa.demos.sofashop.inv.order.OrderInconsistentException;
import org.objectweb.dsrg.sofa.demos.sofashop.store.client.InventoryDataHelper;
import org.objectweb.dsrg.sofa.demos.sofashop.store.client.gui.reports.ReportsPane;

/** {@link SwingWorker} extension for {@link Order orders} registration.  
  * 
  * @author Ondrej Cerny
  *
  */
public class OrderRegistryWorker extends SwingWorker<Integer, OrderTreeNode> {
	
	/** Logger. */
	private static Logger logger = Logger.getLogger(OrderRegistryWorker.class);
	
	/** Tree that displays orders.
	  * 
	  * <p>
	  * Used to initiate UI update after changes.
	  * </p>
	  */
	private JTree tree;

	/** Tree model used from tree that displays the orders.
	  * 
	  * <p>
	  * The {@link Order orders} failed to register will be added back
	  * to this model.
	  * </p>
	  */
	private TreeModel treeModel;
	/** {@link Order Orders} for registration.
	  */
	private Collection<Order> orders;
	/** Show reports for successfully submitted orders?
	  *  
	  */
	private ReportsPane reportsPane;
	
	/** Creates this.
	  *
	  * @param orders {@link Order orders} to register
	  * @param tree {@link JTree tree} for orders view
	  * @param rootNode root {@link MutableTreeNode node} 
	  */
	protected OrderRegistryWorker(JTree tree, TreeModel treeModel, ReportsPane reports) {
		this.treeModel = treeModel;
		this.tree = tree;
		this.reportsPane = reports;
		try {
			this.orders = getOrdersFromTreeModel(treeModel);
		} catch (Throwable t) {
			logger.error(
					"An error occured when getting orders from tree model.",
					t);
			this.orders = null;
		}
		updateUI();
 	}
	/** Creates this and bind it to given tree model.
	  * 
	  * @param treeModel
	  */
	public OrderRegistryWorker(TreeModel treeModel) {
		this(null, treeModel, null);
	}
	/** Creates this and binds it to given tree.
	  * 
	  * @param treeModel
	  */
	public OrderRegistryWorker(JTree tree) {
		this(tree, tree.getModel(), null);
	}
	/** Creates this and binds it to given tree.
	  * 
	  * @param tree tree
	  * @param reports {@link ReportsPane reports} component
	  * 	for showing submitted orders (if null, no reports will be shown)
	  */
	public OrderRegistryWorker(JTree tree, ReportsPane reports) {
		this(tree, tree.getModel(), reports);
	}	
	/** Registers orders using {@link InventoryDataHelper}.
	  * 
	  * @return number of {@link Order} orders that failed registrations 
	  * @see javax.swing.SwingWorker#doInBackground()
	  */
	@Override
	protected Integer doInBackground() throws Exception {
		int failed = 0; // number of failed registrations
		if (orders != null) {
			for(Order o: orders) {
				String code = null;
				try {
					code = InventoryDataHelper.saveOrder(o);
				} catch (OrderInconsistentException oie) {
					logger.info("OrderInconsistentException:" + 
							oie.getMessage());
				} catch (InventoryUnreachableException iue) {
					logger.info("Inventory unreachable when saving orders. " +
							"InventoryUnreachableException: " + iue.getMessage());
				}
				if (code == null) {
					publish(createTreeFromOrder(o));
					++failed;
				} else if (reportsPane != null) {
					// initiate showing the report
					logger.trace("Initiating showing order report for code " + code);
					reportsPane.showOrderReport(code, false);
				}
			}
			
		} else {
			failed = -1;
		}
		return new Integer(failed);
	}
	
	/** Adds order tree nodes to tree model.
	  * 
	  * @param orderTN {@link OrderTreeNode order tree node}.
	  */
	@Override
	protected void process(List<OrderTreeNode> orderTNs) {
		if (orderTNs != null) {
			Object root = treeModel.getRoot();
			MutableTreeNode tmpMTN = null;
			if (root != null && root instanceof MutableTreeNode) {
				tmpMTN = (MutableTreeNode)root;
			}
			if (tmpMTN != null) {
				for (OrderTreeNode oTN: orderTNs) {
					tmpMTN.insert(oTN, tmpMTN.getChildCount());
				}
			} else {
				// fatal error
				logger.error("Failed adding uncommited order back to tree model.");
			}
		}	        
	}
	
	/** Traverses trough the given {@link TreeModel tree model},
	  * extracts orders and returns them in collection.
	  * 
	  * @param treeModel {@link TreeModel tree model}
	  * @return collection of {@link Order orders}
	  * @throws ClassCastException when the tree model is not compatible
	  * @throws NullPointerException when the tree model is not consistent
	  */
	private static Collection<Order> getOrdersFromTreeModel(TreeModel treeModel)
	throws ClassCastException, NullPointerException {
		OrderTreeNode rootTN = (OrderTreeNode) treeModel.getRoot();
		List<Order> result = new LinkedList<Order>();		
		int orderCount = rootTN.getChildCount();
		
		for(int i = 0; i < orderCount; ++i) {
			OrderTreeNode orderOTN = (OrderTreeNode) rootTN.getChildAt(i);
			Order o = orderOTN.getOrder();
			if (o == null) {
				throw new NullPointerException("Null order.");
			}
			List<OrderItem> oiList = o.getOrderItems();
			if (oiList == null) {
				oiList = new LinkedList<OrderItem>();
			}
			int orderItemCount = orderOTN.getChildCount();
			for(int j = 0; j < orderItemCount; ++j) {
				OrderTreeNode orderItemOTN = (OrderTreeNode)
					orderOTN.getChildAt(j);
				OrderItem oi = orderItemOTN.getOrderItem();
				if (oi == null) {
					throw new NullPointerException("Null order item.");
				}
				oiList.add(oi);
			}
			o.setOrderItems(oiList);
			logger.trace("Extracted order with " + oiList.size() +
					" order items.");
			result.add(o);
		}
		rootTN.removeAllChildren();
		return result;
	}
	
	/** Creates order tree (hierarchy of {@link OrderTreeNode order tree nodes})
	  * from {@link Order order} o. 
	  * 
	  * @param o {@link Order order}.
	  * @return {@link OrderTreeNode order tree node} with children
	  */
	private static OrderTreeNode createTreeFromOrder(Order o) {
		OrderTreeNode result = null;
		if (o != null) {
			result = new OrderTreeNode(o);
			List<OrderItem> orderItems = o.getOrderItems();
			if (orderItems != null) {
				for (OrderItem oi : orderItems) {
					result.add(new OrderTreeNode(oi)); 
				}
			} else {
				logger.error("Malformed order - no order items.");
			}
		} else {
			logger.error("Null order can not form order tree.");
		}
		return result;
	}
	
	/** Calls {@link JTree#updateUI()} on {@link #tree}.
	  */
	private void updateUI() {
		if (tree != null) {
			tree.updateUI();
		}
	}
}
