/* $Id: ReportsPane.java 775 2009-10-23 00:49:10Z ondrac $
 *
 * SofaShop demo for SOFA 2.0 component system
 * Copyright (C) 2009  DSRG, Charles University in Prague
 *                     http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.objectweb.dsrg.sofa.demos.sofashop.store.client.gui.reports;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;

import org.apache.log4j.Logger;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.InvalidReportParameterException;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.Report;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportDataProcessingException;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameter;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportParameterLabeledValue;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.ReportPrint;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.util.RemovableReportContainer;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.util.ReportPaneFactory;
import org.objectweb.dsrg.sofa.demos.sofashop.rep.common.util.ReportParameterIntputContainerPane;
import org.objectweb.dsrg.sofa.demos.sofashop.store.client.InventoryDataHelper;

/**
 * @author Ondrej Cerny
 *
 */
public class ReportsPane extends JPanel implements ActionListener, ItemListener, Runnable, RemovableReportContainer {
	private static final String SHOW_REPORT_ACTION_COMMAND = "s";
	
	private LinkedList<JPanel> reportsToBeShown =  new LinkedList<JPanel>();
	
	/** Logger. */
	private static Logger logger = Logger.getLogger(ReportsPane.class);
	
	/** Default constructor. */
	public ReportsPane() {
		init();
		refreshAvailableReports();
	}
	
	/** Refills combo-box for report selection with actual report names.
	  *  
	  */
	public void refreshAvailableReports() {
		
		JComboBox c = getReportsCombo();
		c.setSelectedIndex(-1);
		c.setEnabled(false);
		
		SwingWorker<Set<String>, Object> sw = new SwingWorker<Set<String>, Object>() {
			@Override
			protected Set<String> doInBackground() throws Exception {
				return InventoryDataHelper.getReportNames();
			}
			@Override
			protected void done() {
				Set<String> names = null;
				try {
					names = get();
				} catch (Throwable t) {
					logger.info("Error occurder when getting " +
							"available reports (" + t.getClass().getName() +
							": " + t.getMessage());
				}
				JComboBox combo = getReportsCombo();
				if (names != null && names.size() > 0) {
					logger.trace("Adding " + names.size() + " items to " +
							"reports combo box.");
					combo.setModel(
							new DefaultComboBoxModel(names.toArray()));
				} else {
					logger.trace("No item to be added to " +
							"reports combo box.");					
					combo.setModel(
							new DefaultComboBoxModel());
				}
				combo.setSelectedIndex(-1);
				combo.setEnabled(true);
			}			
		};
		
		sw.execute();
	}
	
	
	/** Returns rendered order report.
	  *
	  * <p>
	  * It is allowed to call this method from any thread,
	  * so it should not block for long time.
	  * </p>
	  * 
	  * @return null if and only if the order has not been found
	  */
	public JPanel getOrderReportPane(String orderCode) {
		JPanel rPane = null;		
		try {
			Report orderReport = InventoryDataHelper.getReportForName("Order");
			if (orderReport != null) {
				List<ReportParameter> pList = orderReport.getParameters();
				if (pList == null) {
					logger.trace("Report has null parameters!");
				} else {
					for(ReportParameter p : pList) {
						if (String.class.equals(p.getType())) {
							logger.trace("Setting report parameter.");
							p.setValue(new ReportParameterLabeledValue(orderCode, orderCode));
						}
					}
					ReportPrint rp = orderReport.getPrint(pList);
					if (rp != null) {
						logger.trace("Obtained report print.");				
						rPane = ReportPaneFactory.createReporPane(rp);
				
					} else {
						logger.trace("Obtained null report print.");
					}
				}
			}
		} catch (ReportDataProcessingException rdpe) {
			logger.trace("Report execution error: " + rdpe.getMessage());
		} catch (InvalidReportParameterException irpe) {
			logger.trace("Report execution error: " + irpe.getMessage());			
		} catch (Throwable t) {
			logger.trace("Error when report processing:", t);
		}
		return rPane;
	}	
	/** Shows order report. 
	  * 
	  * <p>
	  * It is allowed to call this method from any thread,
	  * so it should not block for long time.
	  * </p>
	  * 
	  * @param orderCode
	  * @param inNewWindow
	  */
	public void showOrderReport(String orderCode, boolean inNewWindow) {
		JPanel rPane = getOrderReportPane(orderCode);
		JPanel closablePane = rPane != null 
			?ReportPaneFactory.createClosableReportPaneFromExistingReportPane(
					this, rPane)
			:null;
		if (closablePane != null) {
			logger.trace("Enqueuing report for display and calling the displayer.");
			enqueueReport(closablePane);
			SwingUtilities.invokeLater(this);
		} else {
			logger.info("Report rendering failed.");
		}
	}
	
	/** Enqueues report's {@link JPanel} to be shown.
	  * 
	  */
	public void enqueueReport(JPanel p) {
		synchronized (this) {
			reportsToBeShown.add(p);
		}
	}
	
	/** Returns {@link JPanel} of the first report (not jet displayed).
 	  * 
	  * @return {@link JPanel}
	  */
	public JPanel dequeueReport() {
		JPanel result = null;
		synchronized (this) {
			result = reportsToBeShown.poll();
		}
		return result;
	}
	
	/** Fills the report container with reports from queue reports prepared to be shown. 
	  * 
	  * <p>
	  * Note that this method must run just within the event dispatcher thread.
	  * </p>
	  *
	  * @see java.lang.Runnable#run()
	  */
	@Override
	public void run() {
		JPanel report = null;
		while ((report = dequeueReport()) != null) {
			logger.trace("Displaying report!");
			getReportContainer().add(report);
		}
		getReportContainer().updateUI();
	}

	/** Removes given report pane from the container.
	  *
	  * @param reportPane report to be removed
	  * @see org.objectweb.dsrg.sofa.demos.sofashop.rep.common.util.RemovableReportContainer#removeReport(javax.swing.JPanel)
	  */
	@Override
	public void removeReport(JPanel reportPane) {
		getReportContainer().remove(reportPane);
		this.updateUI();
	}

	public ReportsPane getMe() {
		return this;
	}
	
	//
	// listeners:
	//
	/**
	  * 
	  */
	@Override
	public void actionPerformed(ActionEvent e) {
		
		if (SHOW_REPORT_ACTION_COMMAND.equals(e.getActionCommand())) {
			setControlsEnabled(false);
			logger.trace("Show selected report - showing the report...");
			SwingWorker<String, Object> sw = new SwingWorker<String, Object> () {

				@Override
				protected String doInBackground() throws Exception {
					logger.trace("Running report picker...");
					Report r = getReportParametersPane().getReport();
					if (r != null) {
						ReportPrint rp = null;
						try {
							rp = r.getPrint(
									getReportParametersPane().
										getCurrentParameters());
						} catch (InvalidReportParameterException irpe) {
							logger.trace("Report execution error: " + irpe.getMessage());
							return irpe.getMessage();
						} catch (ReportDataProcessingException rdpe) {
							logger.trace("Report execution error: " + rdpe.getMessage());
						}
						JPanel report = 
							ReportPaneFactory.createClosableReportPane(
									getMe(), rp);
						if (report != null) {
							logger.trace("Enqueuing report to display.");
							enqueueReport(report);
						} else {
							logger.trace("Repor rendering result is null.");
						}
					} else {
						logger.trace("Null report to be shown.");
					}
					logger.trace("Scheduling report displayer to be invoked later.");
					SwingUtilities.invokeLater(getMe());
					return null;
				}
				@Override
				protected void done() {
					String message = null;
					try  {
						message = get();
					} catch (Throwable t) {
						logger.trace("Error when fetching message from report rendering :" +
								t.getClass().getName() + ": " + t.getMessage());
					}
					if (message != null) {
						JOptionPane.showMessageDialog(
								getMe(), message, "Report execution", JOptionPane.WARNING_MESSAGE);
					}
					setControlsEnabled(true);
				}
			};
			
			sw.execute();
		}
	}

	@Override
	public void itemStateChanged(ItemEvent e) {
		if (e.getStateChange() == ItemEvent.SELECTED) {
			logger.trace("Report selection changed - SELECTED");
			Object o = e.getItem();
			if (o != null && o instanceof String) {
				final String name = (String)o;
				logger.trace("report '" + name + "' selected");
				setControlsEnabled(false);
				
				SwingWorker<Report, Object> sw = new SwingWorker<Report, Object> () {

					@Override
					protected Report doInBackground() throws Exception {
						return InventoryDataHelper.getReportForName(name); 
					}
					@Override
					protected void done() {
						Report r = null;
						try {
							r = get();
						} catch (Throwable t) {
							logger.trace("Error occured when getting result of " +
									"InventoryDataHelper.getReportForName('"+ name + "'): "
									+ t.getClass().getName() + ": " + t.getMessage());
						}
						if (r != null) {
							getReportParametersPane().setReport(r);
						} else {
							logger.trace("InventoryDataHelper.getReportForName('" + name + "') == null");
						}
						setControlsEnabled(true);
					}					
				};
				
				sw.execute();
			}
		}
	}
	
	private void setControlsEnabled(boolean val) {
		getReportsCombo().setEnabled(val);
		getShowReportButton().setEnabled(val);
		getReportParametersPane().setEnabled(val);
	}
	
	//
	// boring swing stuff:
	//
	
	private JPanel reportContainer = null;
	private JButton showReportButton = null;
	private JComboBox reportsCombo = null;
	private ReportParameterIntputContainerPane reportParametersPane = null;
	
	
	private void init() {
		
		setLayout(new GridBagLayout());
		JScrollPane p2 = new JScrollPane(getReportParametersPane());
		JPanel p3 = new JPanel();
		p3.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		
		c = new GridBagConstraints();
		c.weighty = 0;
		c.weightx = 0;		
		c.gridy = 0;
		c.gridx = 0;
		c.anchor = GridBagConstraints.NORTHEAST;
		c.fill = GridBagConstraints.BOTH;		
		p3.add(getReportsCombo(), c);
		c = new GridBagConstraints();
		c.weighty = 0;
		c.weightx = 0;		
		c.gridy = 1;
		c.gridx = 0;
		c.anchor = GridBagConstraints.NORTHEAST;
		c.fill = GridBagConstraints.HORIZONTAL;		
		p3.add(getShowReportButton(), c);
		
		c = new GridBagConstraints();
		c.weighty = 0;
		c.weightx = 1;
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.NORTHWEST;		
		c.fill = GridBagConstraints.BOTH;
		add(p2, c);
		c = new GridBagConstraints();
		c.weighty = 0;
		c.weightx = 0;
		c.gridx = 1;
		c.gridy = 0;
		c.ipady = 150;
		c.anchor = GridBagConstraints.NORTHEAST;				
		c.fill = GridBagConstraints.BOTH;
		add(p3, c);
		
		
		JScrollPane sp = new JScrollPane(getReportContainer());
		c = new GridBagConstraints();
		c.weighty = 1;
		c.gridy = 1;
		c.gridx = 0;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.BOTH;
		add(sp, c);		
	}
	public JPanel getReportContainer() {
		if (reportContainer == null) {
			reportContainer = new JPanel();
			reportContainer.setLayout(new BoxLayout(reportContainer, BoxLayout.Y_AXIS));			
		}
		return reportContainer;
	}
	public JButton getShowReportButton() {
		if (showReportButton == null) {
			showReportButton = new JButton("Show");
			showReportButton.setActionCommand(SHOW_REPORT_ACTION_COMMAND);
			showReportButton.addActionListener(this);
			
		}
		return showReportButton;
	}
	public JComboBox getReportsCombo() {
		if (reportsCombo == null) {
			reportsCombo = new JComboBox();			
			reportsCombo.addItemListener(this);
		}
		return reportsCombo;
	}
	
	public ReportParameterIntputContainerPane getReportParametersPane() {
		if (reportParametersPane == null) {
			reportParametersPane = new ReportParameterIntputContainerPane();
		}
		return reportParametersPane;
	}
}
